import spread_bolid as bolid
from ..entities import Funit
from ..providers import Provider


class BolidProvider(Provider):
    def_state = None

    def __init__(self, id: int, number: int, **kwargs):
        super(BolidProvider, self).__init__(id, **kwargs)
        self._number = number
        self._status = self.SUSPICIOUS
        self._state = None

    @staticmethod
    def parse_args(attributes: dict, **kwargs) -> dict:
        return {'number': attributes['number']}

    @property
    def binding(self) -> int:
        return self._number

    def request(self, cmd: bolid.commands.Command, *args, sig=None, **kwargs):
        self.on_request_answer(cmd, self.manager.make_request(cmd, *args, **kwargs), sig=sig)

    def get_info(self):
        pass

    def on_event(self, event: bolid.events.Event):
        raise NotImplementedError()

    def get_guard(self, sig=None):
        raise NotImplementedError()

    def set_guard(self, value, sig=None):
        raise NotImplementedError()


class BolidArea(BolidProvider):
    def check_valid(self):
        self.get_guard()

        validity = self.INVALID if self.get_value(Guard, None) is None else self.VALID
        self.set_valid(validity)

    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type in (Guard, Triggered, StatusCode):
            return False

        return super(BolidArea, self).is_repaired_funit(funit_type)

    def handle_status(self, code, sig=None):
        self.on_update(Guard, code == 24, sig=sig)
        self.on_update(StatusCode, code)
        self.on_update(Triggered, code in [3, 19, 37, 40, 44, 58, 77, 79, 118])

    def on_request_answer(self, cmd, response, *args, sig=None):
        if isinstance(cmd, bolid.commands.ReadAreaState):
            if response is None:
                self.set_valid(self.INVALID)
            else:
                self.set_valid(self.VALID)
                self.handle_status(response >> 8, sig=sig)

    def on_event(self, event: bolid.events.Event):
        if self._number == event.area:
            self.handle_status(event.code)

    def get_guard(self, sig=None):
        self.request(bolid.commands.ReadAreaState(self._number), sig=sig)

    def set_guard(self, value, sig=None):
        self.request(bolid.commands.SetAreaState(self._number, 24 if value else 109))


class BolidFireSensor(BolidArea):
    pass


class BolidIntruderSensor(BolidArea):
    pass


class BolidRelay(BolidProvider):
    def is_repaired_funit(self, funit_type) -> bool:
        if funit_type == On:
            return False

        return super(BolidRelay, self).is_repaired_funit(funit_type)

    def on_event(self, event: bolid.events.Event):
        if self._number == event.relay_index:
            self.on_update(On, bool(event.relay_state))

    def on_request_answer(self, cmd, response, *args, sig=None):
        if isinstance(cmd, bolid.commands.ReadGroupRelay):
            self.on_update(On, response[0], sig=sig)

    def get_guard(self, sig=None):
        self.request(bolid.commands.ReadGroupRelay(self._number, count=1), sig=sig)

    def set_guard(self, value, sig=None):
        self.request(bolid.commands.SetRelayState(self._number, value))


class BoildFunit(Funit):
    entity_type = BolidProvider
    get_func = None
    set_func = None

    @property
    def entity(self) -> BolidProvider:
        return super(BoildFunit, self).entity

    def set(self):
        if self.set_func:
            getattr(self.entity, self.set_func.__name__)(self.set_value, **self.params, sig=self.signature)

    def get(self):
        if self.get_func:
            getattr(self.entity, self.get_func.__name__)(**self.params, sig=self.signature)


BolidProvider.funit_class = BoildFunit


class Guard(BoildFunit):
    entity_type = BolidArea
    value_type = bool
    set_func = BolidProvider.set_guard
    get_func = BolidProvider.get_guard


class On(Guard):
    entity_type = BolidRelay


class BrightnessLevel(Guard):
    entity_type = BolidRelay


class StatusCode(BoildFunit):
    value_type = int


class Triggered(BoildFunit):
    value_type = bool
