import copy
import spread_dali as dali

from spread_core.mqtt.spread.topic import State, Set
from spread_core.mqtt.spread.variable import Variable
from spread_core.tools.signature import CommandSignature
from . import Enginery, _EngineryFunit
from .light import BrightnessLevel
from ..recipe import Recipe

LUMINOSITY_INGREDIENT = 'Luminosity'
PRESENCE_INGREDIENT = 'Presence'
ACTUATION_INGREDIENT = 'Actuation'
BUTTON_INGREDIENT = 'Button'

ACTION_OFF = 'Off'
ACTION_MIN_LEVEL = 'MinLevel'
ACTION_MAX_LEVEL = 'MaxLevel'
ACTION_LAST_LEVEL = 'LastLevel'
ACTION_LEVEL = 'Level'
ACTION_SCENE = 'Scene'
ACTION_UP = 'Up'
ACTION_DOWN = 'Down'
ACTION_STEP_UP = 'StepUp'
ACTION_ON_AND_STEP_UP = 'OnAndStepUp'
ACTION_STEP_DOWN = 'StepDown'
ACTION_STEP_DOWN_AND_OFF = 'StepDownAndOff'
EMULATE_OCCUPANCY = 'EmulateOccupancy'
EMULATE_VACANCY = 'EmulateVacancy'

METHOD_SCENE = 'Scene'
METHOD = 'method'

TT_COMBO = 'Combo'
TT_PRESENCE = 'Presence'
TT_LUMINOSITY = 'Luminosity'
TT_IDLE = 'Idle'

LA_SCENE = 255

HANDLING_BUTTON = 'ButtonHandling'
HANDLING_PRESENCE = 'PresenceHandling'


class LightingArea(Enginery):
    def __init__(self, _id: int, **kwargs):
        super(LightingArea, self).__init__(_id, **kwargs)
        self.buttons = {}
        self.actuators = {}
        self.presences = {}
        self.luminosities = {}
        self.presence_values = dict()
        self.luminosity_values = dict()
        self.tuning = 0
        self.hold_times = {}
        self.lum_ignored = False
        self.last_button_events = {}

    @property
    def is_on(self): return self.get_value(On, False) and not self.get_value(Pause, False)

    def is_dumped(self, key) -> bool:
        return key not in (Presence, Luminosity, Pause) and key not in Profiles.KEYS

    def publish_state(self, funit_type, value, delay=0, **kwargs):
        super(LightingArea, self).publish_state(funit_type, value, delay, **kwargs)

        if funit_type in Profiles.KEYS:
            try:
                current_profile = self.values[CurrentProfile]
                profiles = self.values[Profiles]
                if profiles[current_profile].get(funit_type) != value:
                    profiles = {
                        **profiles,
                        current_profile: {
                            **profiles[current_profile],
                            funit_type: value
                        }
                    }
                    self.publish_state(Profiles, profiles)
            except KeyError:
                pass

    def set_recipe(self, ingredient_type, recipe):
        if ingredient_type == ACTUATION_INGREDIENT:
            self.actuators[recipe.id] = recipe

        elif ingredient_type == PRESENCE_INGREDIENT:
            self.presences[recipe.id] = recipe

        elif ingredient_type == LUMINOSITY_INGREDIENT:
            self.luminosities[recipe.id] = recipe

        elif ingredient_type == BUTTON_INGREDIENT:
            self.buttons[recipe.id] = recipe

        else:
            raise ValueError('unknown ingredient type: "%s"' % ingredient_type)

    @property
    def is_light_bonded(self):
        for recipe_id in self.actuators:
            if self.actuators[recipe_id].is_bonded:
                return True

        return False

    @property
    def is_any_actuator_is_on(self):
        for recipe_id in self.actuators:
            if self.actuators[recipe_id].level:
                return True

        return False

    def provider_command(self, recipe: Recipe, funit_type, value, params=None, sig=None, session=None):
        if self.adapter:
            if sig is None and session:
                sig = CommandSignature(funit_type=funit_type, session_id=session)
            self.adapter.publish(Set(recipe.get_address(funit_type)), Variable(value, params=params, sig=sig))

    def get_info(self):
        for funit_type, def_value in (
                (On, False),
                (Discovery, False),
                (CurrentProfile, CurrentProfile.DEFAULT),
                (Presence, None),
                (Luminosity, None),
                (Pause, False),
                (Buttons, []),
                (Profiles, {profile_name: self.get_profile(profile_name) for profile_name in CurrentProfile.variants})
        ):
            self.publish_state(funit_type, self.get_value(funit_type, def_value))

        self.load_profile(self.get_value(CurrentProfile), hardset=False)

    def get_value(self, funit_type, def_value=None):
        if funit_type in Profiles.KEYS:
            current_profile = self.get_value(CurrentProfile)
            return self.get_value(Profiles, {}).get(current_profile, {}).get(funit_type, def_value)

        return super(LightingArea, self).get_value(funit_type, def_value)

    def get_profile(self, key) -> dict:
        return {
            str(TuningType): TT_COMBO,
            str(OccupancyLevel): 100,
            str(VacancyLevel): 0,
            str(TargetLuminosity): 0,
            str(Hysteresis): 10,
            str(TuningSpeed): 500,
            str(OccupancyAction): ACTION_MAX_LEVEL,
            str(OccupancyScene): 2,
            str(VacancyAction): ACTION_OFF,
            str(VacancyScene): 1,
            str(HoldTime): 600,
            **self.get_value(Profiles, {}).get(key, {})
        }

    def load_profile(self, profile_name, hardset=True):
        """
        Активация профиля. При отсутствии указанного профиля генерируется и применяется дефолтный

        :param profile_name: Индекс профиля, строка вида "ProfileN"
        :param hardset: true - нужна установка targetHoldTime в датчики
        :return:
        """
        self.publish_state(CurrentProfile, profile_name)
        profile = self.get_value(Profiles)[profile_name]
        for funit_type, value in profile.items():
            if funit_type not in ('%sLevels' % OccupancyScene, '%sLevels' % VacancyScene):
                self.publish_state(funit_type, value)

        if hardset:
            HoldTime(self, Variable(self.get_value(HoldTime))).set()

    def load_scene(self, scene_type, sig=None, session=None):
        """
        Включение сцены для всех свеильников в LA. В случае,
        если индекс сцены равен LA_SCENE - сцена читается из LA

        :param scene_type: значение из ["OccupancyScene", "VacancyScene"]
                            или индекс из [0, 15] для сцен девайсов
                            или индекс из ["C1", "C2"] для кастомных сцен (OccupancySceneLevels, VacancySceneLevels)
        :param sig: Сигнатура команды (в случае вызова по команде извне)
        :param session: Идентификатор сессии (в случае вгутреннего вызова)
        :return:
        """

        if scene_type == "C1":
            scene = 'OccupancySceneLevels'

        elif scene_type == "C2":
            scene = 'VacancySceneLevels'

        elif scene_type in (OccupancyScene, VacancyScene):
            index = self.get_value(scene_type)
            if index == 255:
                scene = f'{scene_type}Levels'
            else:
                scene = index
        else:
            assert isinstance(scene_type, int)
            scene = scene_type

        if isinstance(scene, int):
            self.load_scene_index(scene, sig=sig, session=session)
        else:
            self.load_custom_scene(scene, sig=sig, session=session)

    def load_scene_index(self, index, sig=None, session=None):
        assert index in range(16), "Unavailable scene index: %s" % index
        self.set_level(index, METHOD_SCENE, sig=sig, session=session)

    def load_custom_scene(self, scene_name: str, sig=None, session=None):
        """
        Включение кастомной сцены
        :param scene_name: OccupancySceneLevels или OccupancySceneLevels
        """
        for item in self.get_value(scene_name, []):
            self.provider_command(self.actuators[item['id']], BrightnessLevel, item['level'], sig=sig, session=session)

    def save_scene(self, scene_type):
        """
        Сохранение сцены для всех свеильников в LA.
        В случае, если индекс сцены равен LA_SCENE - сцена записываеися в LA
        Иначе - в светильник

        :param scene_type: "OccupancyScene", "VacancyScene" или её индекс из спецификации конечного устройства
        :return:
        """

        scene_levels = [
            {'id': recipe.id, 'level': recipe.level}
            for recipe in self.actuators.values()
            if isinstance(recipe, Recipe) and recipe.level is not None
        ]

        if scene_type == OccupancyScene or scene_type == VacancyScene:
            index = self.get_value(scene_type)
        else:
            index = scene_type

        if index == LA_SCENE:
            self.publish_state('%sLevels' % scene_type, scene_levels)
        else:
            for item in scene_levels:
                self.provider_command(
                    self.actuators[item['id']],
                    'SceneLevelsRaw',
                    [{'index': index, 'value': item['level']}]
                )

    def on_state(self, topic: State, variable: Variable):
        """ Функция обработки состояний провайдеров """

        funit = topic.entity_address.funit_type

        if funit == On:
            if topic.entity_address.id in self.luminosities:
                if variable.value is False or variable.invalid:
                    self.luminosity_values.pop(topic.entity_address.id, None)

                self.check_luminosity()

            elif topic.entity_address.id in self.presences:
                if variable.value is True:
                    self.presence_values.setdefault(topic.entity_address.id, None)
                else:
                    self.presence_values.pop(topic.entity_address.id, None)
            else:
                return
        elif funit == 'CurrentPresence':
            self.on_presence_event(topic.entity_address.id, variable)
        elif funit == 'CurrentLuminosity':
            self.on_luminosity_event(topic.entity_address.id, variable)
        elif funit == 'LastEvent':
            self.on_button_event(topic.entity_address.id, variable)

    def on_luminosity_event(self, sensor_id, variable: Variable):
        if variable.invalid:
            return

        val = int(variable.value)

        if self.luminosity_values.get(sensor_id) != val:
            self.luminosity_values[sensor_id] = val
            self.check_luminosity()

    def on_presence_event(self, provider_id, variable: Variable):
        if variable.invalid:
            return

        val = bool(variable.value)

        if val:
            self.lum_ignored = False

        if self.presence_values.get(provider_id) != val:
            self.presence_values[provider_id] = val
            self.check_presence()

    def on_button_event(self, provider_id, variable: Variable):
        if variable.invalid:
            return

        if provider_id not in self.buttons:
            return

        event = variable.value

        # ищем пресет для сработавшей кнопки
        for pair in self.get_value(Buttons, []):
            if pair.get('index') == list(self.buttons.keys()).index(provider_id):
                preset = self.adapter.button_presets.get(pair.get('preset'), None)
                assert isinstance(preset, ButtonPreset)
                if preset and preset.handlers:
                    index = 0

                    if preset.is_toggle and preset.has_alt:
                        alt_index = self.is_button_event_registered(provider_id, event)

                        if alt_index:
                            index = self.calculate_button_handler_index(event, preset)

                        self.register_button_event(provider_id, event)

                    if index < len(preset.handlers):
                        # ищем сработавшее событие
                        for action in preset.handlers[index]:
                            if action.event == event:
                                self.play_button_action(action)

                                # прерываем цикл поиска события
                                break

                # прерываем цикл поиска пресета
                break

    def is_button_event_registered(self, provider_id, event) -> bool:
        """
        Вычисление необходимости использования
        альтернативного обработчика по принципу очерёдности
        """

        event_stack = self.last_button_events.get(provider_id, [])

        if event == dali.device.events.ButtonEvent.LONG_PRESS_REPEAT:
            return dali.device.events.ButtonEvent.LONG_PRESS_START in event_stack

        return event in event_stack

    def register_button_event(self, provider_id, event):
        """ Регистрация алтернативно обработанного события """

        self.last_button_events.setdefault(provider_id, [])
        event_stack = self.last_button_events.setdefault(provider_id, [])

        if event == dali.device.events.ButtonEvent.LONG_PRESS_STOP:
            if dali.device.events.ButtonEvent.LONG_PRESS_START in event_stack:
                event_stack.remove(dali.device.events.ButtonEvent.LONG_PRESS_START)

        elif event != dali.device.events.ButtonEvent.LONG_PRESS_REPEAT:
            if event in event_stack:
                event_stack.remove(event)
            else:
                event_stack.append(event)

    def calculate_button_handler_index(self, event, preset) -> int:
        """ Выбор индекса обработчика ольтернативно обрабатываемого события кнопки """

        # если такое событие описно во втором массиве обработчиков
        if event in [act.event for act in preset.handlers[1]]:
            # если к LA не привязан ни один светильник или включен хотябы один
            if not self.is_light_bonded or self.is_any_actuator_is_on:
                # вычисление индекса handler`а по принципу невыключенности
                return 1

        return 0

    def play_button_action(self, action):
        _value = None
        _method = None
        _action = action.command

        if _action == ACTION_SCENE:
            self.load_scene(action.scene, session=HANDLING_BUTTON)
        elif _action == ACTION_LEVEL:
            _value = action.level
        elif _action == ACTION_MIN_LEVEL:
            _method = dali.gear.general.RecallMinLevel.__name__
            _value = 1
        elif _action == ACTION_MAX_LEVEL:
            _method = dali.gear.general.RecallMaxLevel.__name__
            _value = 100
        elif _action == ACTION_LAST_LEVEL:
            _method = dali.gear.general.GoToLastActiveLevel.__name__
            _value = 50
        elif _action == ACTION_OFF:
            _method = dali.gear.general.Off.__name__
            _value = 0
        elif _action in (ACTION_UP, ACTION_DOWN, ACTION_STEP_UP, ACTION_STEP_DOWN,
                         ACTION_ON_AND_STEP_UP, ACTION_STEP_DOWN_AND_OFF):
            _method = _action
            _value = -1
        elif _action == EMULATE_OCCUPANCY:
            self.publish_state(Presence, True)
            self.presence_action(True)
        elif _action == EMULATE_VACANCY:
            self.publish_state(Presence, False)
            self.presence_action(False)

        if _value is not None or _method is not None:
            self.set_level(_value, _method, session=HANDLING_BUTTON)

        if action.la:
            On(self, Variable(action.la == On)).set()

        self.lum_ignored = action.lum_ignore
        if self.lum_ignored:
            self.set_tuning(0)

    def check_presence(self, renew=False):
        if len(self.presence_values) == 0:
            return

        _total_presence = True in self.presence_values.values()

        # Если в списке значеие нет ни True ни False, то ничего не делать
        if _total_presence is False and False not in self.presence_values.values():
            return

        if _total_presence == self.get_value(Presence) and not renew:
            return

        self.publish_state(Presence, _total_presence)
        if self.is_on and self.get_value(TuningType) in [TT_COMBO, TT_PRESENCE]:
            self.presence_action(_total_presence)

    def presence_action(self, presence):
        _value = None
        _method = None

        if presence:
            _action = self.get_value(OccupancyAction)

            if _action == ACTION_SCENE:
                self.load_scene(OccupancyScene, session=HANDLING_PRESENCE)
            elif _action == ACTION_MAX_LEVEL:
                _method = dali.gear.general.RecallMaxLevel.__name__
                _value = 100
            elif _action == ACTION_LEVEL:
                _value = self.get_value(OccupancyLevel, 0)
            else:  # _action == ACTION_LAST_LEVEL
                _method = dali.gear.general.GoToLastActiveLevel.__name__
                _value = 50

        else:
            if self.get_value(TuningType) == TT_COMBO:
                self.set_tuning(0)

            _action = self.get_value(VacancyAction)

            if _action == ACTION_MIN_LEVEL:
                _method = dali.gear.general.RecallMinLevel.__name__
                _value = 1
            elif _action == ACTION_LEVEL:
                _value = self.get_value(VacancyLevel, 0)
            elif _action == ACTION_SCENE:
                self.load_scene(VacancyScene, session=HANDLING_PRESENCE)
            else:  # _action == OFF
                _method = dali.gear.general.Off.__name__
                _value = 0

        if _value is not None:
            self.set_level(_value, _method, session=HANDLING_PRESENCE)

    def check_luminosity(self, **kwargs):
        if len(self.luminosity_values) == 0:
            self.set_tuning(0)
            self.publish_state(Luminosity, None)
            return

        avg_lum = round(sum(self.luminosity_values.values()) / len(self.luminosity_values))
        if avg_lum == self.get_value(Luminosity) and not kwargs:
            return

        self.publish_state(Luminosity, avg_lum)

        if kwargs.get(TuningType, self.get_value(TuningType)) == TT_COMBO:
            if not self.get_value(Presence) or not self.is_on:
                self.set_tuning(0)
                return
        elif kwargs.get(TuningType, self.get_value(TuningType)) != TT_LUMINOSITY or not self.is_on:
            self.set_tuning(0)
            return

        if self.is_on and not self.lum_ignored:
            self.luminosity_action(
                avg_lum,
                target=kwargs.get(TargetLuminosity),
                hysteresis=kwargs.get(Hysteresis)
            )

    def luminosity_action(self, avg_lum, target=None, hysteresis=None):
        if target is None:
            target = self.get_value(TargetLuminosity, 0)

        if target:
            if hysteresis is None:
                hysteresis = self.get_value(Hysteresis, 0)

            _dif_rate = round((target - avg_lum) / target, 2)
            if abs(_dif_rate) <= hysteresis / 100:
                _dif_rate = 0

            self.set_tuning(_dif_rate)

    def set_level(self, level: int, method=None, sig=None, session=None):
        for recipe in self.actuators.values():
            self.provider_command(recipe, BrightnessLevel, level, params={METHOD: method}, sig=sig, session=session)

    def set_tuning(self, value):
        if self.tuning != value and self.tuning * value <= 0:
            self.tuning = value
            for recipe in self.actuators.values():
                self.provider_command(recipe, 'Tuning', value, params={'speed': self.get_value(TuningSpeed) / 1000})


class _LAFunit(_EngineryFunit):
    entity_type = LightingArea
    algorithm_dependent = False

    @property
    def entity(self) -> LightingArea: return self._entity

    def get(self):
        if self._variable.session_id:
            self.entity.publish_reply(self._variable.session_id, self.entity.get_value(self.funit_type), self._variable.key)

    def set(self):
        self.entity.publish_state(self.funit_type, self.set_value)
        if self.algorithm_dependent:
            self.entity.check_presence(renew=self.funit_type in (Pause, On, TuningType))
            self.entity.check_luminosity(**{self.funit_type: self.set_value})


LightingArea.funit_class = _LAFunit


class On(_LAFunit):
    value_type = bool
    algorithm_dependent = True


class TuningType(_LAFunit):
    variants = [TT_COMBO, TT_PRESENCE, TT_LUMINOSITY, TT_IDLE]
    algorithm_dependent = True


class Discovery(_LAFunit):
    value_type = bool

    def set(self):
        super(Discovery, self).set()

        for recipe in self.entity.actuators.values():
            self.entity.provider_command(recipe, 'Discovery', self.set_value, sig=self.signature)

        for recipe in self.entity.presences.values():
            self.entity.provider_command(recipe, 'Discovery', self.set_value, sig=self.signature)

        for recipe in self.entity.luminosities.values():
            self.entity.provider_command(recipe, 'Discovery', self.set_value, sig=self.signature)


class OccupancyLevel(_LAFunit):
    variants = range(0, 255)


class VacancyLevel(_LAFunit):
    variants = range(0, 255)


class TargetLuminosity(_LAFunit):
    algorithm_dependent = True


class Hysteresis(_LAFunit):
    algorithm_dependent = True


class TuningSpeed(_LAFunit):
    algorithm_dependent = True


class OccupancyAction(_LAFunit):
    variants = [ACTION_LAST_LEVEL, ACTION_MAX_LEVEL, ACTION_LEVEL, ACTION_SCENE]


class OccupancyScene(_LAFunit):
    variants = list(range(0, 16)) + [255, 'C1', 'C2']


class VacancyAction(_LAFunit):
    variants = [ACTION_OFF, ACTION_MIN_LEVEL, ACTION_LEVEL, ACTION_SCENE]


class VacancyScene(_LAFunit):
    variants = list(range(0, 16)) + [255, 'C1', 'C2']


class Presence(_LAFunit):
    def set(self): raise self.NotImplemented()


class Luminosity(_LAFunit):
    def set(self): raise self.NotImplemented()


class Pause(_LAFunit):
    algorithm_dependent = True
    value_type = bool


class CurrentProfile(_LAFunit):
    DEFAULT = 'Profile1'
    algorithm_dependent = True
    variants = ['Profile%s' % i for i in range(1, 6)]

    def set(self):
        self.entity.load_profile(self.set_value)
        super(CurrentProfile, self).set()


class HoldTime(_LAFunit):
    def set(self):
        super(HoldTime, self).set()
        for recipe in self.entity.presences.values():
            assert isinstance(recipe, Recipe)
            self.entity.provider_command(recipe, 'EngineryHoldTime', self.set_value, sig=self.signature)


class _SceneEditor(_LAFunit):
    scene = None
    def get(self): raise self.NotImplemented()
    def set(self): self.entity.save_scene(self.scene)


class VacancySceneSave(_SceneEditor):
    scene = VacancyScene


class OccupancySceneSave(_SceneEditor):
    scene = OccupancyScene


class _LoadScene(_LAFunit):
    scene = None
    def get(self): raise self.NotImplemented()
    def set(self): self.entity.load_scene(self.scene, sig=self.signature)


class VacancySceneLoad(_LoadScene):
    scene = VacancyScene


class OccupancySceneLoad(_LoadScene):
    scene = OccupancyScene


class Buttons(_LAFunit):
    value_type = list

    @property
    def set_value(self):
        if self._set_value is None:
            buttons = super(Buttons, self).set_value
            for button in buttons:
                if not isinstance(button['index'], int):
                    raise ValueError('"index" required as int')

                if not isinstance(button['preset'], int):
                    raise ValueError('"preset" required as int')

            self._set_value = buttons

        return self._set_value


class Profiles(_LAFunit):
    algorithm_dependent = True
    value_type = dict

    KEYS = {TuningType, OccupancyLevel, VacancyLevel, TargetLuminosity, Hysteresis, TuningSpeed,
            OccupancyAction, OccupancyScene, VacancyAction, VacancyScene, HoldTime,
            '%sLevels' % OccupancyScene, '%sLevels' % VacancyScene}

    def set(self):
        if self.entity.get_value(CurrentProfile) in self.set_value:
            for funit_type, value in self.set_value[self.entity.get_value(CurrentProfile)].items():
                if funit_type in ('OccupancySceneLevels', 'VacancySceneLevels'):
                    continue
                _funit = _LAFunit.of(funit_type)(self.entity, Variable(value))
                if _funit.set_value != self.entity.get_value(funit_type):
                    if funit_type == HoldTime:
                        _funit.set()
                    else:
                        # Funit.set() не выполняется, чтобы не выполнялось действие при флаге algorithm_dependent
                        self.entity.publish_state(funit_type, _funit.set_value)

        self._set_value = self.full_value
        super(Profiles, self).set()

    @property
    def full_value(self):
        profiles = copy.deepcopy(self.entity.get_value(Profiles, {}))
        for profile_name, profile_data in super(Profiles, self).set_value.items():
            profile_name = CurrentProfile(self.entity, Variable(profile_name)).set_value
            profiles[profile_name].update({
                str(key): profile_data.pop(key)
                for key in self.KEYS
                if key in profile_data and (
                        key in ('OccupancySceneLevels', 'VacancySceneLevels')
                        or _LAFunit.of(key)(self.entity, Variable(profile_data[key])).set_value != 'q1w2e3r4t5y6'
                )
            })

            if len(profile_data) > 0:
                raise ValueError('unexpected keys %s' % profile_data.keys())

        return profiles


class VirtualPresence(_LAFunit):
    value_type = bool

    def get(self): raise self.NotImplemented()
    def set(self): self.entity.presence_action(self.set_value)


class VirtualLuminosity(_LAFunit):
    value_type = int

    def get(self): raise self.NotImplemented()
    def set(self): self.entity.luminosity_action(self.set_value)


LightingArea.gettable_units = [
    On,
    TuningType,
    Discovery,
    OccupancyLevel,
    VacancyLevel,
    TargetLuminosity,
    Hysteresis,
    TuningSpeed,
    OccupancyAction,
    OccupancyScene,
    VacancyAction,
    VacancyScene,
    OccupancySceneSave,
    OccupancySceneLoad,
    VacancySceneSave,
    VacancySceneLoad,
    Presence,
    Luminosity,
    Pause,
    CurrentProfile,
    Profiles,
    HoldTime,
    Buttons
]

LightingArea.settable_units = [
    On,
    TuningType,
    Discovery,
    OccupancyLevel,
    VacancyLevel,
    TargetLuminosity,
    Hysteresis,
    TuningSpeed,
    OccupancyAction,
    OccupancyScene,
    VacancyAction,
    VacancyScene,
    OccupancySceneSave,
    OccupancySceneLoad,
    VacancySceneSave,
    VacancySceneLoad,
    Pause,
    CurrentProfile,
    Profiles,
    HoldTime,
    Buttons,
    VirtualPresence,
    VirtualLuminosity
]


class ButtonPreset:
    def __init__(self, _id: int):
        self.id = _id
        self.is_toggle = False
        self.name = f'ButtonPreset({_id})'
        self.handlers = [[], []]

    def update(self, data):
        self.is_toggle = data.get('isToggle', False)
        self.name = data.get('name', self.name)
        for i in range(0, 2):
            if len(data['handlers']) > i:
                self.handlers[i] = [ButtonAction(action) for action in data['handlers'][i]]
            else:
                self.handlers[i] = []

        return self

    @property
    def has_alt(self): return len(self.handlers) > 1 and len(self.handlers[1]) > 0

    def __str__(self):
        return '<%s>%s' % (self.id, ';'.join([','.join([str(a) for a in h])for h in self.handlers]))


class ButtonAction:
    __slots__ = ['event', 'command', 'setLightingArea', 'stopTuning', 'level', 'scene']

    def __init__(self, data):
        if data['event'] not in dali.device.events.ButtonEvent.STATES.values():
            raise ValueError('указано некорректное значение для "event": %s' % data['event'])

        if data['command'] not in set(VacancyAction.variants
                                      + OccupancyAction.variants
                                      + [
                                          ACTION_UP,
                                          ACTION_DOWN,
                                          ACTION_STEP_UP,
                                          ACTION_STEP_DOWN,
                                          ACTION_ON_AND_STEP_UP,
                                          ACTION_STEP_DOWN_AND_OFF,
                                          EMULATE_OCCUPANCY,
                                          EMULATE_VACANCY,
                                      ]):
            raise ValueError('указано некорректное значение для "command": %s' % data['command'])

        if data['command'] == ACTION_SCENE and (data['scene'] not in range(0, 16) and data['scene'] not in ["C1", "C2"]):
            raise ValueError('scene ожидается как 0..15 или C1 или C2')

        if data['command'] == ACTION_LEVEL and data['level'] not in range(0, 256):
            raise ValueError('"level" ожидается как 0..255')

        if data.get('setLightingArea', None) not in ['On', 'Off', None]:
            raise ValueError('"setLightingArea" не ожидается или ожидается как "On"/"Off"')

        if not isinstance(data.get('stopTuning', False), bool):
            raise ValueError('"stopTuning" ожидается как bool')

        for attr in self.__slots__:
            if attr in data:
                setattr(self, attr, data[attr])

    @property
    def la(self): return self.setLightingArea if hasattr(self, 'setLightingArea') else None

    @property
    def lum_ignore(self): return self.stopTuning if hasattr(self, 'stopTuning') else False

    def __str__(self):
        return '%s->%s' % (
            self.event,
            f'{self.command}({self.level})' if self.command == 'Level' else
            f'{self.command}({self.scene})' if self.command == 'Scene' else
            f'{self.command}'
        )

    def __repr__(self):
        return self.__str__()
