# import json
import re
from spread_core.bam.validation.base import corr_par, corr_req_par, int_float2d
import spread_core.bam.validation.constraints as constr


class EngineryValidator():
    def __init__(self, data):
        if not isinstance(data, dict):
            raise ValueError('Wrong data structure (dict needed)')
        if not corr_req_par('type', str, data) or data['type'] not in constr.ENG_CONSTR:
            raise ValueError('Wrong type')
        if not corr_req_par('id', int, data) or data['id'] < 1:
            raise ValueError('Wrong id')
        # key must be GUID
        if not corr_req_par('key', str, data) \
                or not re.match(r'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', data['key']):
            raise ValueError('Wrong key')
        if not corr_par('name', str, data):
            raise ValueError('Wrong name')
        if 'recipe' in data:
            if not isinstance(data['recipe'], dict):
                raise ValueError('Wrong recipe structure (dict needed)')
            self.Recipe(data)
        if 'attributes' in data:
            if not isinstance(data['attributes'], dict):
                raise ValueError('Wrong attributes structure (dict needed)')
            self.Attributes(data['type'], data['attributes'])
        else:
            # Check for type attributes needed
            self.Attributes(data['type'])
        if not corr_par('label', str, data):
            raise ValueError('Wrong label')

    def Recipe(self, data):
        # Рецепт должен соответствовать типу объекта Enginery
        if not corr_req_par('type', str, data['recipe']) \
                or data['recipe']['type'] not in constr.ENG_CONSTR[data['type']]:
            raise ValueError('Wrong recipe type')
        if 'ingredients' in data['recipe']:
            if not isinstance(data['recipe']['ingredients'], list):
                raise ValueError('Wrong ingredients structure (list needed)')

            for ingr in data['recipe']['ingredients']:
                if not isinstance(ingr, dict):
                    raise ValueError('Wrong ingredient structure (dict needed)')
                if not corr_req_par('type', str, ingr):
                    raise ValueError('Wrong recipe ingredient type')
                if not corr_req_par('index', int, ingr) or ingr['id'] < 0:
                    raise ValueError('Wrong recipe ingredient index')
                if not corr_req_par('id', int, ingr) or ingr['id'] < 1:
                    raise ValueError('Wrong recipe ingredient id')
            self.Ingredients(data)

    def Ingredients(self, data):
        """ Extended ingredients verification """
        # В рецепте любой допустимый ингредиент может встречаться один раз или отсутствовать, но
        # если в Constraints ингредиент оканчивается на [], то он может встречаться несколько раз или отсутствовать
        ingredients_types_list = [ingr['type'] for ingr in data['recipe']['ingredients']]
        constraints_ingredients_types_list = constr.ENG_CONSTR[data['type']][data['recipe']['type']].copy()
        try:
            for ingr in ingredients_types_list:
                if (ingr + '[]') in constraints_ingredients_types_list:
                    pass
                else:
                    del constraints_ingredients_types_list[ingr]
        except:
            raise ValueError('Wrong recipe ingredients')

    @classmethod
    def Attributes(cls, eng_type: str, attr=None):
        if not isinstance(eng_type, str):
            raise ValueError('Wrong enginery type')

        error = f"Wrong {eng_type} attribute"
        if eng_type == 'Panel':
            if not attr:
                raise ValueError(f"{error}s")
            if not corr_req_par('type', str, attr):
                raise ValueError(f"{error} type")
            if not corr_par('serial', str, attr):
                raise ValueError(f"{error} serial")
            if 'sip' in attr:
                if not corr_req_par('address', str, attr['sip']):
                    raise ValueError(f"{error} sip address")
                if not corr_par('password', str, attr['sip']):
                    raise ValueError(f"{error} sip password")

        elif eng_type == 'Camera':
            if not attr:
                raise ValueError(f"{error}s")
            if not corr_req_par('url', str, attr):
                raise ValueError(f"{error} url")
            if not corr_req_par('codec', str, attr) or attr['codec'] not in ('Auto', 'Mxpeg', 'Mjpeg'):
                raise ValueError(f"{error} codec")

        elif eng_type == 'DoorPhone':
            if 'sip' in attr:
                if not corr_req_par('address', str, attr['sip']):
                    raise ValueError(f"{error} sip address")
                if 'dtmfCommands' in attr['sip']:
                    if not isinstance(attr['sip']['dtmfCommands'], list):
                        raise ValueError(f"{error} sip dtmfCommands structure (list needed)")

                    for com in attr['sip']['dtmfCommands']:
                        if not isinstance(com, dict):
                            raise ValueError(f"{error} sip dtmfCommand structure (dict needed)")

                        if not corr_req_par('tag', str, com) or com['tag'] not in ('Unlock'):
                            raise ValueError(f"{error} sip dtmfCommand tag")
                        # value must contain only DTMF simbols
                        if not corr_req_par('value', str, com) or not re.match(r'[0-9A-D*#]+$', com['value']):
                            raise ValueError(f"{error} sip dtmfCommand value")

            if 'video' in attr:
                if not corr_req_par('source', str, attr['video']) or attr['video']['source'] not in ('Camera'):
                    raise ValueError(f"{error} video source")
                if not corr_par('id', int, attr['video']) or attr['video']['id'] < 1:
                    raise ValueError(f"{error} video id")

        elif eng_type in ('Curtain', 'Blind'):
            if not attr:
                raise ValueError(f"{error}s")
            if not corr_req_par('direction', str, attr) or attr['direction'] not in ('CloseDown', 'CloseLeft', 'CloseRight', 'CloseCenter'):
                raise ValueError(f"{error} direction")

        elif eng_type == 'DimmingLight':
            if not attr:
                raise ValueError(f"{error}s")
            if not corr_req_par('positionCount', int, attr) or attr['positionCount'] < 0:
                raise ValueError(f"{error} positionCount")
            if not corr_req_par('lowerLimit', int, attr) or attr['lowerLimit'] < 0:
                raise ValueError(f"{error} lowerLimit")
            if not corr_req_par('upperLimit', int, attr) or attr['upperLimit'] < 0:
                raise ValueError(f"{error} upperLimit")
            if not corr_req_par('power', int, attr) or attr['power'] < 0:
                raise ValueError(f"{error} power")

        elif eng_type in ('SwitchingLight', 'RgbLight', 'RgbwLight'):
            if not attr:
                raise ValueError(f"{error}s")
            if not corr_req_par('power', int, attr) or attr['power'] < 0:
                raise ValueError(f"{error} power")

        elif eng_type == 'DynamicLight':
            if not attr:
                raise ValueError(f"{error}s")
            if 'scenarios' not in attr:
                raise ValueError(f"{error} scenarios")

            if not isinstance(attr['sip']['scenarios'], list):
                raise ValueError(f"{error} sip scenarios structure (list needed)")

            for scen in attr['scenarios']:
                if not isinstance(scen, dict):
                    raise ValueError(f"{error} sip scenario structure (dict needed)")

                if not corr_req_par('index', int, scen) or scen['index'] < 1:
                    raise ValueError(f"{error} scenario index")
                if not corr_req_par('number', int, scen) or scen['number'] < 1:
                    raise ValueError(f"{error} scenario number")
                if not corr_req_par('name', str, scen):
                    raise ValueError(f"{error} scenario name")
                if not corr_req_par('color', str, scen) or not re.match(r'#[0-9A-F]{6}$', scen['color']):
                    raise ValueError(f"{error} scenario color")

        elif eng_type == 'Thermoregulator':
            if 'target' in attr:
                if 'min' not in attr['target'] \
                        or not int_float2d(attr['target']['min']) \
                        or attr['target']['min'] < 223 \
                        or attr['target']['min'] > 573:
                    raise ValueError(f"{error} target min")

                if 'max' not in attr['target'] \
                        or not int_float2d(attr['target']['max']) \
                        or attr['target']['max'] < 223 \
                        or attr['target']['max'] > 573:
                    raise ValueError(f"{error} target max")

                d = attr['target']['max'] - attr['target']['min']
                if d < 0:
                    raise ValueError(f"{error} target min must be less then max")

                if 'step' not in attr['target'] \
                        or not int_float2d(attr['target']['step']) \
                        or attr['target']['step'] < 0.01 \
                        or attr['target']['step'] > d:
                    raise ValueError(f"{error} target step")

            if 'presets' in attr:
                if not isinstance(attr['presets'], list):
                    raise ValueError(f"{error} presets structure (list needed)")
                values = ['Manual', 'Auto', 'Protection', 'Day', 'Night', 'Standby', 'Comfort']
                for pre in attr['presets']:
                    if pre in values:
                        values.remove(pre)
                    else:
                        raise ValueError(f"{error} presets")

            if 'modes' in attr:
                if not isinstance(attr['modes'], list):
                    raise ValueError(f"{error} modes structure (list needed)")
                values = ['Manual', 'Auto', 'Cooling', 'Heating', 'Drying', 'Ventilation', 'HeatPlus', 'Stop']
                for pre in attr['modes']:
                    if pre in values:
                        values.remove(pre)
                    else:
                        raise ValueError(f"{error} modes")

            if 'fanSpeedModes' in attr:
                if not isinstance(attr['fanSpeedModes'], list):
                    raise ValueError(f"{error} fanSpeedModes structure (list needed)")
                values = ['Manual', 'Auto', 'High', 'Medium', 'Low', 'Top']
                for pre in attr['fanSpeedModes']:
                    if pre in values:
                        values.remove(pre)
                    else:
                        raise ValueError(f"{error} fanSpeedModes")

            if 'louverModes' in attr:
                if not isinstance(attr['louverModes'], list):
                    raise ValueError(f"{error} louverModes structure (list needed)")
                values = ['Manual', 'Auto', 'Horizontal', 'Angle30', 'Angle45', 'Angle60', 'Vertical']
                for pre in attr['louverModes']:
                    if pre in values:
                        values.remove(pre)
                    else:
                        raise ValueError(f"{error} louverModes")

        elif eng_type == 'Scenario':
            if not attr:
                raise ValueError(f"{error}s")
            if not corr_req_par('title', str, attr):
                raise ValueError(f"{error} title")
            if not corr_req_par('icon', int, attr) or attr['icon'] < 0:
                raise ValueError(f"{error} icon")

        elif eng_type == 'MeetingRoom':
            if not attr:
                raise ValueError(f"{error}s")
            if not corr_req_par('displayName', str, attr):
                raise ValueError(f"{error} displayName")

        elif eng_type == 'TunableWhiteLight':
            if not attr:
                raise ValueError(f"{error}s")
            if not corr_req_par('power', int, attr) or attr['power'] < 0:
                raise ValueError(f"{error} power")
            if not corr_req_par('positionCount', int, attr) or attr['positionCount'] < 0:
                raise ValueError(f"{error} positionCount")
            if not corr_req_par('lowerLimit', int, attr) or attr['lowerLimit'] < 0:
                raise ValueError(f"{error} lowerLimit")
            if not corr_req_par('upperLimit', int, attr) or attr['upperLimit'] < 0:
                raise ValueError(f"{error} upperLimit")
            if not corr_req_par('warmestColor', int, attr) or attr['warmestColor'] < 0:
                raise ValueError(f"{error} warmestColor")
            if not corr_req_par('coolestColor', int, attr) or attr['coolestColor'] < 0:
                raise ValueError(f"{error} coolestColor")

        elif attr:
            raise ValueError(f"{error}s")
