# import json
import re
from spread_core.bam.validation.base import corr_par, corr_req_par
import spread_core.bam.validation.constraints as constr


class LocationValidator:
    def __init__(self, data):
        if not isinstance(data, dict):
            raise ValueError('Wrong data structure (dict needed)')
        if not corr_req_par('id', int, data) or data['id'] < 1:
            raise ValueError('Wrong id')
        # key must be GUID
        if not corr_req_par('key', str, data) \
                or not re.match(r'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', data['key']):
            raise ValueError('Wrong key')
        if not corr_par('label', str, data):
            raise ValueError('Wrong label')
        if not corr_req_par('name', str, data) or not data['name']:
            raise ValueError('Wrong name')
        if not corr_par('title', str, data):
            raise ValueError('Wrong title')

        if not corr_req_par('navigations', list, data):
            raise ValueError('Wrong navigations structure (list needed)')

        for num in data['navigations']:
            if not isinstance(num, int) or num < 1:
                raise ValueError('Wrong navigations')

        if not corr_req_par('arrangements', list, data) or not data['arrangements']:
            raise ValueError('Wrong arrangements structure (not null list needed)')

        for arr in data['arrangements']:
            self.Arrangement(arr)

        if not corr_req_par('models', list, data):
            raise ValueError('Wrong models structure (list needed)')

        for model in data['models']:
            if not isinstance(model, dict):
                raise ValueError('Wrong model structure (dict needed)')
            if not corr_req_par('id', int, model) or model['id'] < 1:
                raise ValueError('Wrong model id')

            if not corr_req_par('transitions', list, model):
                raise ValueError('Wrong model transitions structure (list needed)')

            for tran in model['transitions']:
                if not isinstance(tran, dict):
                    raise ValueError('Wrong model transitions structure (dict needed)')
                if not corr_req_par('locationID', int, tran) or tran['locationID'] < 1:
                    raise ValueError('Wrong model transition locationID')
                if not corr_req_par('sameLevel', bool, tran):
                    raise ValueError('Wrong model transition sameLevel')

            if not corr_req_par('controls', list, model):
                raise ValueError('Wrong model controls structure (list needed)')

            for cont in model['controls']:
                if not isinstance(cont, dict):
                    raise ValueError('Wrong model controls structure (dict needed)')
                if not corr_req_par('engineryID', int, cont) or cont['engineryID'] < 1:
                    raise ValueError('Wrong model control engineryID')
                if not corr_req_par('clickable', bool, cont):
                    raise ValueError('Wrong model control clickable')
                if 'controlArrangement' in cont:
                    try:
                        self.Arrangement(cont['controlArrangement'])
                    except ValueError as ex:
                        raise ValueError(f'Wrong model control controlArrangement ({ex})')

        if 'statusBar' in data:
            if not isinstance(data['statusBar'], list):
                raise ValueError('Wrong statusBar structure (list needed)')

            for num in data['statusBar']:
                if not isinstance(num, int) or num < 1:
                    raise ValueError('Wrong statusBar')

    def Arrangement(self, arr):
        if not isinstance(arr, dict):
            raise ValueError('Wrong arrangement structure (dict needed)')
        if not corr_req_par('title', str, arr) or not arr['title']:
            raise ValueError('Wrong arrangement title')

        if not corr_req_par('position', dict, arr):
            raise ValueError('Wrong arrangement position structure (dict needed)')
        if not corr_req_par('x', float, arr['position']):
            raise ValueError('Wrong arrangement position x')
        if not corr_req_par('y', float, arr['position']):
            raise ValueError('Wrong arrangement position y')
        if not corr_req_par('z', float, arr['position']):
            raise ValueError('Wrong arrangement position z')

        if not corr_req_par('center', dict, arr):
            raise ValueError('Wrong arrangement center structure (dict needed)')
        if not corr_req_par('x', float, arr['center']):
            raise ValueError('Wrong arrangement center x')
        if not corr_req_par('y', float, arr['center']):
            raise ValueError('Wrong arrangement center y')
        if not corr_req_par('z', float, arr['center']):
            raise ValueError('Wrong arrangement center z')

        if not corr_req_par('models', list, arr):
            raise ValueError('Wrong arrangement models structure (list needed)')

        for model in arr['models']:
            if not isinstance(model, dict):
                raise ValueError('Wrong arrangement model structure (dict needed)')
            if not corr_req_par('id', int, model) or model['id'] < 1:
                raise ValueError('Wrong arrangement model id')
            if not corr_req_par('visibility', bool, model):
                raise ValueError('Wrong arrangement model visible')

            if not corr_req_par('position', dict, model):
                raise ValueError('Wrong arrangement model position structure (dict needed)')
            if not corr_req_par('x', float, model['position']):
                raise ValueError('Wrong arrangement model position x')
            if not corr_req_par('y', float, model['position']):
                raise ValueError('Wrong arrangement model position y')
            if not corr_req_par('z', float, model['position']):
                raise ValueError('Wrong arrangement model position z')

        if not corr_req_par('widget', list, arr):
            raise ValueError('Wrong arrangement widget structure (list needed)')

        for widget in arr['widget']:
            if not isinstance(widget, dict):
                raise ValueError('Wrong arrangement widget structure (dict needed)')
            if not corr_req_par('enginery_id', int, widget) or widget['enginery_id'] < 1:
                raise ValueError('Wrong arrangement widget enginery_id')
            if not corr_req_par('subginery_type', str, widget) or widget['subginery_type'] not in constr.SUBG_CONSTR:
                raise ValueError('Wrong arrangement widget subginery_type')
            if not corr_req_par('type', str, widget) or widget['type'] not in constr.SUBG_CONSTR[widget['subginery_type']]:
                raise ValueError('Wrong arrangement widget type')

            if not corr_req_par('title', str, widget) or not widget['title']:
                raise ValueError('Wrong arrangement widget title')

            if not corr_req_par('anchor', str, widget) \
                    or widget['anchor'] not in ("LeftUp", "LeftDown", "RightUp", "RightDown"):
                raise ValueError('Wrong arrangement widget anchor')
            if not corr_req_par('shift_x', str, widget) or widget['shift_x'] < 0:
                raise ValueError('Wrong arrangement widget shift_x')
            if not corr_req_par('shift_y', str, widget) or widget['shift_y'] < 0:
                raise ValueError('Wrong arrangement widget shift_y')

            if not corr_req_par('height', str, widget) or widget['height'] < 1:
                raise ValueError('Wrong arrangement widget height')
            if not corr_req_par('width', str, widget) or widget['width'] < 1:
                raise ValueError('Wrong arrangement widget width')

            # ??? attributes may be here
