import asyncio
import json
import logging
import os
import shutil
import signal
import threading
import zipfile
import zlib

import spread_dali as dali
import spread_lom as lom

from asyncio_mqtt import Client, MqttError

from spread_core import mqtt
from spread_core.tools import settings, utils
from spread_core.bam.engineries import lighting_area as la


def check_la_btn_preset_action(data):
    if data['event'] not in dali.device.events.ButtonEvent.STATES.values():
        raise ValueError('event ожидается как один из %s' % dali.device.events.ButtonEvent.STATES.values())

    if data['command'] not in set(la.VacancyAction.variants
                                  + la.OccupancyAction.variants
                                  + [
                                      la.ACTION_UP,
                                      la.ACTION_DOWN,
                                      la.ACTION_STEP_UP,
                                      la.ACTION_STEP_DOWN,
                                      la.ACTION_ON_AND_STEP_UP,
                                      la.ACTION_STEP_DOWN_AND_OFF,
                                      la.EMULATE_OCCUPANCY,
                                      la.EMULATE_VACANCY,
                                  ]):
        raise ValueError('указано некорректное значение для "command"')

    if data['command'] == la.ACTION_SCENE and (data['scene'] not in range(0, 16) and data['scene'] not in ["C1", "C2"]):
        raise ValueError('scene ожидается как 0..15 или C1 или C2')

    if data['command'] == la.ACTION_LEVEL and data['level'] not in range(0, 256):
        raise ValueError('level ожидается как 0..255')

    if data.get('setLightingArea', None) not in ['On', 'Off', None]:
        raise ValueError('"setLightingArea" не ожидается или ожидается как "On"/"Off"')

    if not isinstance(data.get('stopTuning', False), bool):
        raise ValueError('"stopTuning" ожидается как bool')


def clear_project(mosquitto_db):
    logging.info('mosquitto clean...')

    os.system('sudo service mosquitto stop')

    if os.path.isdir(mosquitto_db):
        shutil.rmtree(mosquitto_db, ignore_errors=True)
    elif os.path.isfile(mosquitto_db):
        os.remove(mosquitto_db)

    os.system('sudo service mosquitto start')
    logging.info('mosquitto cleaned')


async def publish_project(client, zip_path):
    try:
        await asyncio.sleep(3)
    except asyncio.CancelledError:
        # logging.debug('publish_project cancelled')
        pass
    else:
        zip_data = utils.read_file(zip_path, binary=True)

        with zipfile.ZipFile(zip_path, 'r') as input_zip:
            def read_json(file):
                with input_zip.open(file, "r") as fp:
                    return json.load(fp)

            project_id = read_json(mqtt.project.ProjectFile.HEADER + '.json')['header']['project']['id']

            tasks = (
                asyncio.create_task(
                    publish(client, mqtt.project.ProjectFile.ctp(project_id), zip_data)
                ),
                *(
                    asyncio.create_task(
                        publish(client, mqtt.project.ProjectFile(project_id, file), read_json(file))
                    )
                    for file in input_zip.namelist()
                    if file != mqtt.project.ProjectFile.MODELS + '.json'
                ),
            )

        os.system('sudo spread_utils_dashboard fill --all')
        try:
            await asyncio.gather(*tasks)
        except asyncio.CancelledError:
            logging.debug('Ignore cancelling publishing project')
        except Exception as ex:
            logging.exception(ex)

        await publish(client, mqtt.project.Hash(project_id), zlib.crc32(zip_data))
        logging.info('Project updated')


async def publish_local(client, project_id, local_path):
    try:
        await asyncio.sleep(3)
    except asyncio.CancelledError:
        # logging.debug('publish_locals cancelled')
        pass
    else:

        root = os.path.join(local_path, mqtt.project.LOCAL)
        tasks = set()
        for dirpath, dirnames, filenames in os.walk(root):
            if dirpath.endswith(mqtt.project.Cycle.family):
                continue
            for file in filenames:
                tasks.add(asyncio.create_task(publish(
                    client,
                    mqtt.project.Local.of_file(
                        project_id,
                        dirpath.split(root)[1],
                        file.split(os.path.extsep)[0]
                    ),
                    utils.read_file(os.path.join(dirpath, file))
                )))

        try:
            await asyncio.gather(*tasks)
        except asyncio.CancelledError:
            logging.debug('Ignore cancelling publishing project')
        except Exception as ex:
            logging.exception(ex)


def check_hash(zip_path, mosquitto_db, current_hash):
    real_hash = zlib.crc32(utils.read_file(zip_path, binary=True))

    if real_hash != current_hash:
        clear_project(mosquitto_db)


async def check_local(topic: mqtt.project.Preset, data):
    file_path = topic.path(settings.config.get(settings.LOCAL_PATH))

    preset = utils.read_file(file_path, no_raise=True)

    if preset != data:
        if data == '':
            if os.path.exists(file_path):
                os.remove(file_path)
                return
        elif topic.entity_type == la.ButtonPreset.__name__:
            try:
                if data['name'] == '':
                    raise ValueError('"name" не может быть пустым')

                if not isinstance(data.get('isToggle', False), bool):
                    raise ValueError('"isToggle" Ожидается как bool')

                for handler in data.get('handlers', []):
                    for action in handler:
                        check_la_btn_preset_action(action)

            except BaseException as ex:
                logging.debug(ex)

                if data.get('name'):
                    if preset is None:
                        preset = {}

                    preset.update({
                        key: data.get(key, default)
                        for key, default in (
                            ('name', 'AutoName'),
                            ('isToggle', False),
                            ('handlers', []),
                        )})

                    for handler in preset['handlers'].copy():
                        for action in handler.copy():
                            try:
                                check_la_btn_preset_action(action)
                            except:
                                handler.remove(action)

                    return preset
        elif topic.entity_address.entity_type == lom.profiles.DimmingProfile.__name__:
            try:
                lom.profiles.GlobalDimmingProfile.deserialize({'id': topic.entity_id, **data})
            except BaseException as ex:
                logging.debug(ex)
                return preset

        utils.write_file(data, file_path)


async def subscribe(client, topic: mqtt.project.Topic, logged=True):
    await client.subscribe(str(topic))
    if logged:
        logging.debug('subscribed to %s' % topic)


async def publish(client, topic: (str, mqtt.Topic), data, retain=True, logged=True):
    if not isinstance(data, (int, str, bytes, bytearray)):
        data = json.dumps(data)

    if logged:
        logging.debug('[%s] -> %s' % (topic, data if data.__sizeof__() <= 1024 else ('%s bytes' % len(str(data)))))
    await client.publish(str(topic), data, retain=retain)


async def listener():
    PROJECT_ID = settings.config[settings.PROJECT_ID]
    ZIP_PATH = settings.config[settings.CTP_PATH]
    MOSQUITTO_DB = settings.config[settings.MOSQUITTO_PATH]
    LOCAL_PATH = settings.config[settings.LOCAL_PATH]
    RECONNECT_INTERVAL = settings.config.get(settings.RECONNECT_INTERVAL, 3)

    while True:
        try:
            async with Client(
                    hostname=settings.config[settings.BROKER].get(settings.HOST, '127.0.0.1'),
                    port=settings.config[settings.BROKER].get(settings.PORT, 1883),
                    username=settings.config[settings.BROKER].get(settings.LOGIN),
                    password=settings.config[settings.BROKER].get(settings.PASSWORD),
                    client_id='Projector'
            ) as client:

                async with client.filtered_messages(str(mqtt.project.ProjectTopic(PROJECT_ID))) as messages:
                    await asyncio.gather(
                        asyncio.get_event_loop().create_task(subscribe(client, mqtt.project.Hash(PROJECT_ID))),
                        asyncio.get_event_loop().create_task(subscribe(client, mqtt.project.Preset.broadcast(PROJECT_ID)))
                    )

                    project_publisher = asyncio.create_task(publish_project(client, ZIP_PATH))
                    local_publisher = asyncio.create_task(publish_local(client, PROJECT_ID, LOCAL_PATH))
                    async for message in messages:
                        try:
                            topic = mqtt.topic_of(message.topic)
                            data = json.loads(message.payload) if message.payload else ''

                            logging.debug('(%s)[%s] <- %s' % (threading.active_count(), message.topic, data))

                            if isinstance(topic, mqtt.project.Hash):
                                if project_publisher:
                                    project_publisher.cancel()
                                    del project_publisher
                                    project_publisher = None

                                check_hash(ZIP_PATH, MOSQUITTO_DB, data)

                            elif isinstance(topic, mqtt.project.Preset):
                                if local_publisher:
                                    local_publisher.cancel()
                                    del local_publisher
                                    local_publisher = None

                                to_publish = await check_local(topic, data)
                                if to_publish:
                                    await publish(client, message.topic, to_publish)

                                del to_publish
                            del topic, data
                        except BaseException as ex:
                            logging.exception(ex)
        except MqttError as error:
            logging.error(error)
            await asyncio.sleep(RECONNECT_INTERVAL)
        except (KeyboardInterrupt, asyncio.CancelledError):
            print('breaked')
            break
        except BaseException as ex:
            logging.exception(ex)


def run():
    signal.signal(signal.SIGHUP, lambda signum, frame: settings.config.parse())
    try:
        asyncio.run(listener())
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    run()
