from unittest import TestCase

from spread_core.bam.engineries.lighting_area import ButtonPreset, ButtonAction


class TestButtonPreset(TestCase):
    def setUp(self) -> None:
        self.preset = ButtonPreset(1)

    def test_parser_1(self):
        self.preset.update({
            "name": "Test Button",
            "isToggle": True,
            "handlers": [
                [
                    {
                        "event": "DoublePress",
                        "command": "MinLevel"
                    }
                ],
                [
                    {
                        "event": "ShortPress",
                        "command": "Off",
                    }
                ]
            ]
        })

        self.assertEqual(getattr(self.preset, 'name'), 'Test Button')
        self.assertEqual(getattr(self.preset, 'is_toggle'), True)
        self.assertEqual(getattr(self.preset, 'has_alt'), True)

    def test_parser_2(self):
        self.preset.update({
            "isToggle": True,
            "handlers": []
        })

        self.assertEqual(getattr(self.preset, 'name'), 'ButtonPreset(1)')
        self.assertEqual(getattr(self.preset, 'is_toggle'), True)
        self.assertEqual(getattr(self.preset, 'has_alt'), False)

    def test_parser_3(self):
        self.preset.update({
            "name": "Test Button 2",
            "isToggle": True,
            "handlers": [
                [
                    {
                        "event": "DoublePress",
                        "command": "MinLevel"
                    }
                ]
            ]
        })

        self.assertEqual(getattr(self.preset, 'name'), 'Test Button 2')
        self.assertEqual(getattr(self.preset, 'is_toggle'), True)
        self.assertEqual(getattr(self.preset, 'has_alt'), False)


class TestButtonAction(TestCase):
    @staticmethod
    def create_action(data: dict):
        return ButtonAction(data)
    
    def test_action(self):
        action = self.create_action({
            "event": "ShortPress",
            "command": "Off",
            "stopTuning": True,
            "setLightingArea": "On"
        })

        self.assertEqual(action.event, 'ShortPress')
        self.assertEqual(action.command, 'Off')
        self.assertEqual(action.lum_ignore, True)
        self.assertEqual(action.la, 'On')

        action = self.create_action({
            "event": "ShortPress",
            "command": "Off",
        })

        self.assertEqual(action.event, 'ShortPress')
        self.assertEqual(action.command, 'Off')
        self.assertEqual(action.lum_ignore, False)
        self.assertEqual(action.la, None)

    def test_events(self):
        for event in [
            'ButtonReleased', 'ButtonPressed', 'ShortPress', 'DoublePress', 'LongPressStart',
            'LongPressRepeat', 'LongPressStop', 'ButtonFree', 'ButtonStuck'
        ]:
            action = self.create_action({
                "event": event,
                "command": 'MinLevel',
            })

            self.assertEqual(action.event, event)
            self.assertEqual(action.command, 'MinLevel')

    def test_commands(self):
        for command in [
            'Off', 'MinLevel', 'MaxLevel', 'LastLevel', 'Up', 'Down', 'StepUp', 'StepDown', 'OnAndStepUp',
            'StepDownAndOff', 'EmulateOccupancy', 'EmulateVacancy'
        ]:
            action = self.create_action({
                "event": 'ShortPress',
                "command": command,
            })

            self.assertEqual(action.event, 'ShortPress')
            self.assertEqual(action.command, command)

    def test_command_level(self):
        action = self.create_action({
            "event": 'ShortPress',
            "command": 'Level',
            "level": 50
        })

        self.assertEqual(action.event, 'ShortPress')
        self.assertEqual(action.command, 'Level')
        self.assertEqual(action.level, 50)

    def test_command_scene(self):
        for scene in 12, 'C1', 'C2':
            action = self.create_action({
                "event": 'ShortPress',
                "command": 'Scene',
                "scene": scene
            })

            self.assertEqual(action.event, 'ShortPress')
            self.assertEqual(action.command, 'Scene')
            self.assertEqual(action.scene, scene)

    def test_missing_level(self):
        with self.assertRaises(KeyError) as ex:
            self.create_action({
                "event": 'ShortPress',
                "command": 'Level',
            })

        self.assertEqual(ex.exception.args[0], 'level')

    def test_invalid_level(self):
        for level in 'qwe', -1, 256, None:
            with self.assertRaises(ValueError) as ex:
                self.create_action({
                    "event": 'ShortPress',
                    "command": 'Level',
                    "level": level
                })

            self.assertEqual(ex.exception.args[0], '"level" ожидается как 0..255')

    def test_missing_scene(self):
        with self.assertRaises(KeyError) as ex:
            self.create_action({
                "event": 'ShortPress',
                "command": 'Scene',
            })

        self.assertEqual(ex.exception.args[0], 'scene')

    def test_invalid_scene(self):
        for scene in None, 'qwe', -1, 16:
            with self.assertRaises(ValueError) as ex:
                self.create_action({
                    "event": 'ShortPress',
                    "command": 'Scene',
                    "scene": scene
                })

            self.assertEqual(ex.exception.args[0], 'scene ожидается как 0..15 или C1 или C2')

    def test_missing_event(self):
        with self.assertRaises(KeyError) as ex:
            self.create_action({
                "command": "Off",
            })

        self.assertEqual(ex.exception.args[0], 'event')

    def test_invalid_event(self):
        with self.assertRaises(ValueError) as ex:
            self.create_action({
                "event": "qwe",
                "command": "Off",
            })

        self.assertEqual(ex.exception.args[0], 'указано некорректное значение для "event": qwe')

    def test_missing_command(self):
        with self.assertRaises(KeyError) as ex:
            self.create_action({
                "event": "ShortPress",
            })

        self.assertEqual(ex.exception.args[0], 'command')

    def test_invalid_command(self):
        with self.assertRaises(ValueError) as ex:
            self.create_action({
                "event": "ShortPress",
                "command": "qwe",
            })

        self.assertEqual(ex.exception.args[0], 'указано некорректное значение для "command": qwe')
