from unittest import TestCase

from spread_core.bam.generator import generate_subginery
from spread_core.errors.project_errors import ProjectError


class TestSubgineryGeneration(TestCase):
    def _check_subginery_generation(self, provider_type, attributes=None, **attributes_for_check):
        if attributes is None:
            attributes = {'locationID': 998, 'engineries': [1, 2, 3]}

            attributes_for_check['_location_id'] = 998
            attributes_for_check['_enginery_ids'] = [1, 2, 3]

        subginery = generate_subginery(999, provider_type, attributes)

        self.assertEqual(subginery.id, 999)
        self.assertEqual(provider_type, type(subginery).__name__)

        for attr_name in attributes_for_check:
            try:
                self.assertEqual(getattr(subginery, attr_name), attributes_for_check[attr_name])
            except AttributeError as ex:
                self.fail(str(ex))

        return subginery

    def test_unknown_type(self):
        with self.assertRaises(ProjectError) as ex:
            self._check_subginery_generation('SomeNonEexistType', {})

        self.assertEqual(ex.exception.args[0], 'Unknown Subginery type "SomeNonEexistType(999)"')

    def test_missing_location_id(self):
        with self.assertRaises(ProjectError) as ex:
            self._check_subginery_generation('Subginery', {'engineries': []})

        self.assertEqual(ex.exception.args[0], "Subginery(999): parse_args() missing 1 required positional argument: 'locationID'")

    def test_missing_engineries_list(self):
        with self.assertRaises(ProjectError) as ex:
            self._check_subginery_generation('Subginery', {'locationID': 'invalidInt'})

        self.assertEqual(ex.exception.args[0], "Subginery(999): parse_args() missing 1 required positional argument: 'engineries'")

    def test_invalid_location_id(self):
        with self.assertRaises(ProjectError) as ex:
            self._check_subginery_generation('Subginery', {'locationID': 'invalidInt', 'engineries': []})

        self.assertEqual(ex.exception.args[0], "Subginery(999): invalid literal for int() with base 10: 'invalidInt'")

    def test_invalid_engineries_list(self):
        with self.assertRaises(ProjectError) as ex:
            self._check_subginery_generation('Subginery', {'locationID': '123', 'engineries': None})

        self.assertEqual(ex.exception.args[0], "Subginery(999): 'NoneType' object is not iterable")

    def test_lighting(self):
        self._check_subginery_generation("Lighting")

    def test_shading(self):
        self._check_subginery_generation("Shading")

    def test_handling(self):
        self._check_subginery_generation("Handling")

    def test_climate(self):
        self._check_subginery_generation("Climate")
