import copy
import json
import logging
from unittest import TestCase, skip

import spread_core.bam.dali.external_dispatcher as ext
import spread_core.bam.entities as e_lib
import spread_core.bam.managers as m_lib
import spread_core.tools.settings_const as C
import spread_dali as dali
import spread_dali.device.general as device
import spread_dali.gear.general as gear
from spread_core import mqtt
from spread_core.bam.dali.managers import Dali
from spread_core.bam.dali.providers.lights import DaliDimmer, DaliRelay, DaliTunableWhite
from spread_core.bam.dali.providers.sensors import DaliPresenceSensor, DaliLightSensor, DaliPushButton
from spread_core.bam.dali.tuner import TunerLimited
from spread_core.mqtt.spread.variable import Variable
from spread_core.tests import utils
from spread_core.tools.incrementor import Increment
from spread_dali.address import Short, Group, Broadcast, InstanceType, InstanceNumber, InstanceGroup
from spread_dali.device.events import Event

logging.getLogger().setLevel(logging.ERROR)


P_ID = 0
M_ID = 1

DALI_MANAGER = {
    "type": "Dali",
    "id": M_ID,
    "name": "Тестовый объект Dali",
    "attributes": {
        "pollRate": 1000,
        "bus": "Rapida/99/Dali/2/3"
    },
}

CONFIG = {
    "include": [
        "/etc/spread/common.json"
    ],
    "managerID": 1,
    "pollRate": None,
    "useTimestamp": True,
    "communicateTimeout": 1.0,
    "localAdapter": False,
    "levelSyncDelay": 1.0
}


def parse(data):
    pass


class _TestDali(TestCase):
    @staticmethod
    def start_interface_handler(instance):
        pass

    @classmethod
    def setUpClass(cls) -> None:
        dali.interface.can.Bus = utils.TestCanBus
        m_lib.config.args = [None]

    def setUp(self) -> None:
        self.manager = Dali(P_ID, M_ID)
        self.warehouse = self.manager.warehouse

        global parse
        parse = self.manager.parse

        self.manager.set_broker(utils.TestMqttClient())
        parse(DALI_MANAGER)
        self.manager.set_interface()

    def append_provider(self, item, fixture=None, group_ids: list = None):
        if fixture:
            fixture(item)

        self.warehouse.add_provider(item)
        self.warehouse.add_binding(item.binding, item)

        if group_ids:
            for group_id in group_ids:
                self.manager.add_to_group(item, group_id)


class TestDaliInitialization(_TestDali):
    def test_device_count(self):
        self.assertEqual(len(self.warehouse.providers), 0)
        self.assertEqual(len(self.warehouse.bindings), 0)

    def test_name(self):
        self._set_and_parse(C.NAME, "bla bla bla")
        self.assertEqual(self.manager.name, "bla bla bla")

        self._set_and_parse(C.NAME, None, clear=True)
        self.assertEqual(self.manager.name, f'id:{M_ID}')

    def test_bus(self):
        self._set_and_parse(C.BUS, 'bla bla bla', for_attrs=True)
        self.assertEqual(self.manager.bus, 'bla bla bla')

        self._set_and_parse(C.BUS, None, for_attrs=True, clear=True)
        self.assertIsNone(self.manager.bus)

    def test_poll_rate(self):
        self._set_and_parse(C.POLL_RATE, 1234, for_attrs=True)
        self.assertEqual(self.manager.poll_rate, 1.2)
        self.assertEqual(self.manager._def_poll_rate, 1234)

    def test_poll_rate_empty(self):
        self._set_and_parse(C.POLL_RATE, 0, for_attrs=True, clear=True)
        self.assertEqual(self.manager.poll_rate, 0)
        self.assertEqual(self.manager._def_poll_rate, 0)

    def test_empty_data(self):
        try:
            parse(None)
        except TypeError:
            return
        except:
            self.fail("Failed empty data parsing")
        else:
            self.fail("Parsing empty data did not trigger completion!")

    @staticmethod
    def _set_and_parse(key, value, for_attrs=False, clear=False):
        init_data = copy.deepcopy(DALI_MANAGER)
        data_for_change = init_data[C.ATTRIBUTES] if for_attrs else init_data
        if clear:
            data_for_change.pop(key, None)
        else:
            data_for_change[key] = value

        parse(init_data)


class TestDaliConfigurations(_TestDali):
    def set_and_update(self, key, value, clear=False):
        if clear:
            m_lib.config.pop(key, None)
        else:
            m_lib.config[key] = value

        self.manager.update_config()

    def test_send_interface(self):
        self.set_and_update(C.LOCAL_ADAPTER, None, clear=True)
        self.assertIsInstance(self.manager.send_interface, dali.interface.BrokerAdapter)

        self.set_and_update(C.LOCAL_ADAPTER, True)
        self.assertIsInstance(self.manager.send_interface, dali.interface.DaliCanAdapter)

        self.set_and_update(C.LOCAL_ADAPTER, False)
        self.assertIsInstance(self.manager.send_interface, dali.interface.BrokerAdapter)

    def test_poll_rate(self):
        self.set_and_update(C.POLL_RATE, 3210)
        self.assertEqual(self.manager.poll_rate, 3.2)

        self.set_and_update(C.POLL_RATE, None)
        self.assertEqual(self.manager.poll_rate, 1)

        self.set_and_update(C.POLL_RATE, 0, clear=True)
        self.assertEqual(self.manager.poll_rate, 1)

        self.set_and_update(C.POLL_RATE, None)
        self.assertEqual(self.manager.poll_rate, 1)

    def test_level_sync_delay(self):
        self.set_and_update(C.LEVEL_SYNC_DELAY, 0.5)
        self.assertEqual(self.manager.level_sync_delay, 0.5)

    def test_tuning_by_rates(self):
        self.set_and_update(C.TUNING_BY_RATES, True)
        self.assertIs(TunerLimited.UP, gear.Up)
        self.assertIs(TunerLimited.DOWN, gear.Down)

        self.set_and_update(C.TUNING_BY_RATES, False)
        self.assertIs(TunerLimited.UP, gear.StepUp)
        self.assertIs(TunerLimited.DOWN, gear.StepDown)

    def test_communicate_timeout(self):
        self.set_and_update(C.COMMUNICATE_TIMEOUT, 999)
        self.assertEqual(self.manager.send_interface.timeout, 999)


class TestDaliEventsReceiving(_TestDali):
    def test_event_receiver_type_1(self):
        self._test_event(
            DaliPushButton(999, Short(63), InstanceNumber(31)),
            Event(Short(63), InstanceNumber(31), 0)
        )

    def test_event_receiver_type_3(self):
        self._test_event(
            DaliPresenceSensor(999, Short(63)),
            Event(Short(63), InstanceType(3), 0)
        )

    def test_event_receiver_type_4(self):
        self._test_event(
            DaliLightSensor(999, Short(63)),
            Event(Short(63), InstanceType(4), 0)
        )

    @skip
    def test_event_receiver_group_existent(self):
        self._test_event(
            DaliPresenceSensor(2, Group(1)),
            Event(Group(63), InstanceType(3), 0)
        )

    @skip
    def test_event_receiver_group_nonexistent(self):
        self._test_event(
            DaliPushButton(2, Short(1), InstanceNumber(0)),
            Event(Short(63), InstanceType(4), 0)
        )

    def _test_event(self, provider, event):
        provider.on_event = utils.Handler.proceed

        self.append_provider(provider)
        try:
            self.manager.on_event(event)
        except utils.Handler as ex:
            self.assertIs(ex.args[0], event)
        else:
            self.fail(f'Unhandled {event}')


class TestDaliEventIgnoring(_TestDali):
    @staticmethod
    def fixture(provider):
        provider.on_event = utils.Handler.proceed

    def test_event_receiver_no_address(self):
        self._test_event(
            DaliPresenceSensor(1, Short(0)),
            Event(Short(63), InstanceType(3), 0)
        )

    def test_event_receiver_no_type(self):
        self._test_event(
            DaliPresenceSensor(1, Short(0)),
            Event(Short(0), InstanceType(1), 0)
        )

    def test_event_receiver_no_index(self):
        self._test_event(
            DaliPushButton(1, Short(0), InstanceNumber(0)),
            Event(Short(0), InstanceNumber(1), 0)
        )

    def test_event_receiver_no_group(self):
        self._test_event(
            DaliPushButton(1, Short(0), InstanceNumber(0)),
            Event(Group(1), InstanceType(1), 0),
            group_ids=[0]
        )

    def test_event_receiver_no_group_2(self):
        self._test_event(
            DaliPushButton(2, Group(0), InstanceNumber(0)),
            Event(Group(1), InstanceType(1), 0)
        )

    def test_event_receiver_no_instance_type(self):
        self._test_event(
            DaliPushButton(1, Short(0), InstanceNumber(0)),
            Event(InstanceType(31), InstanceNumber(0), 0)
        )

    def test_event_receiver_no_instance_group(self):
        self._test_event(
            DaliPresenceSensor(1, Short(0)),
            Event(InstanceGroup(31), InstanceType(1), 0)
        )

    def _test_event(self, provider, event, group_ids=None):
        self.append_provider(provider, fixture=self.fixture, group_ids=group_ids)
        try:
            self.manager.on_event(event)
        except utils.Handler as ex:
            self.fail(f'Mistakenly handled {ex.args[0]}')


class TestExternalDispatcherSource(_TestDali):
    def simulate_receiving(self, command: dali.command.Command, response=None, *args):
        self.manager.on_external_command(command, response, *args)

    def get_dtr_value(self, dali_type, index):
        return self.manager.get_dtr_value(dali_type, index)

    def test_dali1_dtr0(self):
        self.simulate_receiving(gear.DTR0(123))
        self.assertEqual(123, self.get_dtr_value(1, 0))

    def test_dali1_dtr1(self):
        self.simulate_receiving(gear.DTR1(123))
        self.assertEqual(123, self.get_dtr_value(1, 1))

    def test_dali1_dtr2(self):
        self.simulate_receiving(gear.DTR2(123))
        self.assertEqual(123, self.get_dtr_value(1, 2))

    def test_dali2_dtr0(self):
        self.simulate_receiving(device.DTR0(123))
        self.assertEqual(123, self.get_dtr_value(2, 0))

    def test_dali2_dtr1(self):
        self.simulate_receiving(device.DTR1(123))
        self.assertEqual(123, self.get_dtr_value(2, 1))

    def test_dali2_dtr2(self):
        self.simulate_receiving(device.DTR2(123))
        self.assertEqual(123, self.get_dtr_value(2, 2))

    def test_dali2_dtr10(self):
        self.simulate_receiving(device.DTR1DTR0(12, 34))
        self.assertEqual(12, self.get_dtr_value(2, 1))
        self.assertEqual(34, self.get_dtr_value(2, 0))

    def test_dali2_dtr21(self):
        self.simulate_receiving(device.DTR2DTR1(23, 45))
        self.assertEqual(23, self.get_dtr_value(2, 2))
        self.assertEqual(45, self.get_dtr_value(2, 1))


class TestExternalDispatcherTargets(_TestDali):
    @classmethod
    def setUpClass(cls) -> None:
        super(TestExternalDispatcherTargets, cls).setUpClass()
        cls.SingleHandler = ext.SingleExternalHandler
        cls.GroupHandler = ext.GroupExternalHandler
        cls.BroadcastHandler = ext.BroadcastExternalHandler

    def setUp(self) -> None:
        super(TestExternalDispatcherTargets, self).setUp()
        self.dispatcher = ext.ExternalDispatcher(self.manager)
        self.get_handler = ext.ExternalHandlerAbstract.of
        self.is_sources_command = self.dispatcher.is_for_source

        i = Increment(0)

        for binding_index in range(0, 3):
            self.append_provider(DaliDimmer(i.next(), Short(binding_index)), group_ids=[binding_index % 3])

        for binding_index in range(3, 6):
            self.append_provider(DaliRelay(i.next(), Short(binding_index)), group_ids=[binding_index % 3])

        for binding_index in range(6, 9):
            self.append_provider(DaliTunableWhite(i.next(), Short(binding_index)), group_ids=[binding_index % 3])

        for binding_index in range(3):
            self.append_provider(DaliPresenceSensor(i.next(), Short(binding_index)))
            self.append_provider(DaliLightSensor(i.next(), Short(binding_index)))

        for binding_index in range(3, 6):
            for instance_index in range(3):
                self.append_provider(DaliPushButton(i.next(), Short(binding_index), InstanceNumber(instance_index)))

        self.append_provider(DaliDimmer(i.next(), Group(0)))

    def test_handler_selector(self):
        self.assertRaises(TypeError, self.get_handler, gear.DTR0(0))
        self.assertRaises(TypeError, self.get_handler, device.DTR0(0))

        self.assertIs(self.get_handler(gear.QueryStatus(63)), ext.SingleExternalHandler)
        self.assertIs(self.get_handler(gear.DAPC(Group(15), 0)), ext.GroupExternalHandler)
        self.assertIs(self.get_handler(gear.DAPC(Broadcast(), 0)), ext.BroadcastExternalHandler)

    def test_sources_dali1(self):
        self.assertTrue(self.is_sources_command(gear.DTR0(0)))
        self.assertFalse(self.is_sources_command(gear.DAPC(63, 123)))
        self.assertFalse(self.is_sources_command(gear.QueryStatus(63)))

    def test_sources_dali2(self):
        self.assertTrue(self.is_sources_command(device.DTR0(0)))
        self.assertFalse(self.is_sources_command(device.QueryDeviceStatus(63)))
        self.assertFalse(self.is_sources_command(device.QueryInstanceStatus(63, InstanceNumber(31))))

    def test_targets_dali1_single_existent(self):
        self.append_provider(DaliDimmer(999, Short(63)))
        handler = self.SingleHandler(self.manager, gear.DAPC(63, 123))
        self.check_count(handler, 999)

    def test_targets_dali2_single_existent(self):
        self.append_provider(DaliPresenceSensor(999, Short(63)))
        self.append_provider(DaliLightSensor(998, Short(63)))

        handler = self.SingleHandler(self.manager, device.EnableInstance(63, InstanceNumber(0)))
        self.check_count(handler, 999, 998)

    def test_targets_dali1_single_nonexistent(self):
        handler = self.SingleHandler(self.manager, gear.DAPC(63, 123))
        self.check_count(handler)

    def test_targets_dali2_single_nonexistent(self):
        handler = self.SingleHandler(self.manager, device.EnableInstance(63, InstanceNumber(0)))
        self.check_count(handler)

    def test_targets_dali1_group_existent(self):
        self.append_provider(DaliDimmer(999, Group(15)))

        handler = self.GroupHandler(self.manager, gear.DAPC(Group(15), 123))
        self.check_count(handler, 999)

    def test_targets_dali2_group_existent(self):
        self.append_provider(DaliPresenceSensor(999, Group(15)))
        self.append_provider(DaliLightSensor(998, Group(15)))

        handler = self.GroupHandler(self.manager, device.EnableInstance(Group(15), InstanceNumber(0)))
        self.check_count(handler, 999, 998)

    def test_targets_dali1_group_nonexistent(self):
        self.append_provider(DaliDimmer(999, Short(63)), group_ids=[0, 15])
        self.append_provider(DaliDimmer(998, Short(62)), group_ids=[1, 15])
        self.append_provider(DaliDimmer(997, Short(61)), group_ids=[2, 15])

        handler = self.GroupHandler(self.manager, gear.DAPC(Group(15), 123))
        self.check_count(handler, 999, 998, 997)

    def test_targets_dali2_group_nonexistent(self):
        self.append_provider(DaliPresenceSensor(999, Short(63)), group_ids=[0, 15])
        self.append_provider(DaliLightSensor(998, Short(63)), group_ids=[0, 15])
        self.append_provider(DaliPushButton(997, Short(62), InstanceNumber(0)), group_ids=[1, 15])
        self.append_provider(DaliPushButton(996, Short(62), InstanceNumber(1)), group_ids=[1, 15])

        handler = self.GroupHandler(self.manager, device.EnableInstance(Group(15), InstanceNumber(0)))
        self.check_count(handler, 999, 998, 997, 996)

    def test_targets_dali1_broadcast_existent(self):
        self.append_provider(DaliDimmer(999, Broadcast()))
        handler = self.BroadcastHandler(self.manager, gear.DAPC(Broadcast(), 123))

        self.check_count(handler, 999)

    def test_targets_dali2_broadcast_existent(self):
        self.append_provider(DaliPresenceSensor(999, Broadcast()))
        handler = self.BroadcastHandler(self.manager, device.EnableInstance(Group(15), InstanceNumber(0)))

        self.check_count(handler, 999)

    def test_targets_dali1_broadcast_nonexistent(self):
        handler = self.BroadcastHandler(self.manager, gear.DAPC(Broadcast(), 123))
        self.check_count(handler, *range(1, 10))

    def test_targets_dali2_broadcast_nonexistent(self):
        handler = self.BroadcastHandler(self.manager, device.EnableInstance(Broadcast(), InstanceNumber(0)))
        self.check_count(handler, *range(10, 25))

    def check_count(self, handler, *expected_ids):
        count = len(expected_ids)
        targets = handler.targets

        for _ in range(count):
            try:
                self.assertIn(next(targets).id, expected_ids)
            except StopIteration:
                self.fail(msg=f'Applied items less than {count}')

        with self.assertRaises(StopIteration, msg=f'Applied items more than {count}'):
            next(targets)


class TestCommandReceiver(_TestDali):
    @classmethod
    def setUpClass(cls) -> None:
        super(TestCommandReceiver, cls).setUpClass()
        m_lib.Entity.funit_class = utils.TestFunit
        e_lib.Funit = utils.TestFunit

    def setUp(self) -> None:
        super(TestCommandReceiver, self).setUp()
        self.manager.warehouse = utils.TestStore()
        self.reply_topic = mqtt.spread.topic.Reply(P_ID, 'test_session').__str__()
        self.variable = Variable('testValue', session='test_session')

    def test_entity_address(self):
        address = self.manager.get_address('TestFunit')
        self.assertIsInstance(address, mqtt.spread.address.ManagerAddress)
        self.assertEqual(address.type, "Dali")
        self.assertEqual(address.id, M_ID)
        self.assertEqual(address.funit_type, "TestFunit")

    def test_entity_address_sharp(self):
        address = self.manager.get_address(None)
        self.assertIsInstance(address, mqtt.spread.address.ManagerAddress)
        self.assertEqual(address.type, "Dali")
        self.assertEqual(address.id, M_ID)
        self.assertEqual(address.funit_type, "#")

    def test_invalid_bus(self):
        address = mqtt.spread.address.ProviderAddress(P_ID, "Manager", M_ID, "Provider", 999, "TestFunit")
        self.manager.on_command(mqtt.spread.topic.Set(address), self.variable)

        self.assertEqual(len(self.manager.mqttc.messages), 1)

        topic, data, retain, delay, *_ = self.manager.mqttc.messages[0]
        self.assertEqual(self.reply_topic, topic)
        self.assertEqual(0, delay)
        self.assertFalse(retain)

        data = json.loads(data)

        self.assertTrue(data.get('invalid'))
        self.assertEqual(data.get('error'), "Bus is not ready")

    @skip
    def test_manager_command(self):
        pass

    def test_provider_command(self):
        address = mqtt.spread.address.ProviderAddress(P_ID, "Manager", M_ID, "Provider", 999, "TestFunit")

        self.manager.send_interface.ready = True
        with self.assertRaises(utils.Popup) as popup:
            self.manager.on_command(mqtt.spread.topic.Set(address), self.variable)

        self.assertEqual(popup.exception.args[0], 999)


class TestStateReceiver(_TestDali):
    @staticmethod
    def fixture(provider):
        provider.on_state = utils.Handler.proceed

    def test_state_receiver(self):
        self.append_provider((DaliDimmer(1, Short(0))), fixture=self.fixture)
        address = mqtt.spread.address.ProviderAddress(P_ID, "Dali", M_ID, "DaliDimmer", 1, "SomeFunitOfState")
        topic = mqtt.spread.topic.State(address)
        try:
            self.manager.on_state(topic, Variable('Success'))
        except utils.Handler as ex:
            self.assertEqual(ex.args[0], "SomeFunitOfState")
            self.assertEqual(ex.args[1].value, "Success")
        else:
            self.fail(f'Unhandled state')


@skip
class TestExitHandler(_TestDali):
    pass
