from unittest import TestCase

from spread_core.bam.lom import LomDimmer

from spread_core.bam.dali.providers.lights import DaliDimmer, DaliRelay, DaliTunableWhite
from spread_core.bam.dali.providers.sensors import DaliPresenceSensor, DaliLightSensor, DaliPushButton
from spread_core.bam.knx.providers import KnxDimmer
from spread_core.bam.mercury import MercuryElectricMeter
from spread_core.tools.incrementor import Increment
from spread_core.warehouse.warehouse_interface import HardwareWarehouse
from spread_dali.address import Short, Group, Broadcast, InstanceNumber
from spread_knx.address import Address as KnxAddress


class TestStore(TestCase):
    def append_provider(self, provider):
        self.warehouse.add_provider(provider)
        self.warehouse.add_binding(provider.binding, provider)

    def setUp(self) -> None:
        self.warehouse = HardwareWarehouse()

        i = Increment(0)

        for binding_index in range(0, 3):
            self.append_provider(DaliDimmer(i.next(), Short(binding_index)))

        for binding_index in range(3, 6):
            self.append_provider(DaliRelay(i.next(), Short(binding_index)))

        for binding_index in range(6, 9):
            self.append_provider(DaliTunableWhite(i.next(), Short(binding_index)))

        for binding_index in range(3):
            self.append_provider(DaliDimmer(i.next(), Group(binding_index)))

        for binding_index in range(3):
            self.append_provider(DaliPresenceSensor(i.next(), Short(binding_index)))
            self.append_provider(DaliLightSensor(i.next(), Short(binding_index)))

        for binding_index in range(3, 6):
            short_address = Short(binding_index)
            for instance_index in range(3):
                instance = InstanceNumber(instance_index)
                self.append_provider(DaliPushButton(i.next(), short_address, instance))

    def test_counts(self):
        self.assertEqual(27, len(self.warehouse.providers))

    def test_device(self):
        provider = DaliDimmer(999, Short(63))
        self.add_and_check(provider, Short(63))

    def test_group(self):
        provider = DaliDimmer(999, Group(31))
        self.add_and_check(provider, Group(31))

    def test_broadcast(self):
        provider = DaliDimmer(999, Broadcast())
        self.add_and_check(provider, Broadcast())

    def test_device2(self):
        provider = DaliPresenceSensor(999, Short(63))
        self.add_and_check(provider, Short(63))

    def test_device_mercury(self):
        provider = MercuryElectricMeter(999, 99, {
            "password2": "222222",
            "password1": "111111"
        })

        self.add_and_check(provider, 99)

    def test_device_knx_dimmer(self):
        provider = KnxDimmer(999, KnxAddress('add1', 'add2'))
        self.add_and_check(provider,  KnxAddress('add1', 'add2'))

    def test_device_lom_dimmer(self):
        provider = LomDimmer(999, 'dev_eui', None)
        self.add_and_check(provider,  'dev_eui')

    def add_and_check(self, provider, binding):
        self.append_provider(provider)
        self.assertEqual(28, len(self.warehouse.providers))
        self.assertIs(provider, self.warehouse.get_provider(999))
        self.assertIs(provider, list(self.warehouse.providers_by_address(binding))[0])

    def test_grouping(self):
        provider = DaliDimmer(998, Short(63))

        self.warehouse.include_in_group(2, provider)
        self.assertIn(provider, self.warehouse.groups[2])
        self.assertEqual(1, self.warehouse.group_size(2))

        self.warehouse.exclude_from_group(2, provider)
        self.assertNotIn(2, self.warehouse.groups)
        self.assertEqual(0, self.warehouse.group_size(2))
