from collections import deque
from unittest import TestCase

from spread_core.bam.knx.manager import Knx

from spread_core.bam.dali.managers import Dali
from spread_core.bam.lom.manager import Lom

from spread_dali.frame import BackwardFrame


class ManagerMock:
    def __init__(self):
        self.queue = deque()

    def publish(self, topic, variable, retain=False, delay=0, **kwargs):
        self.queue.append((topic, variable, retain, delay))

    def check_end_of_publishes(self, context: TestCase):
        if self.queue:
            unexpected_publishes = f'Unexpected publishes({len(self.queue)} pcs):\n'
            unexpected_publishes += "\n".join(str(i) for i in self.queue)
            context.fail(unexpected_publishes)

    def check_publish(self, context: TestCase, _topic, _value, _retain=True, _delay=0, params=None, attrs=None):
        if len(self.queue) == 0:
            context.fail('Expected publish is absent: %s' % _topic)
        else:
            topic, variable, retain, delay = self.queue.popleft()
            context.assertEqual(_topic, topic)
            context.assertEqual(_value, variable.value)
            context.assertEqual(_retain, retain)
            context.assertEqual(_delay, delay)

            if params:
                for param_name in params:
                    try:
                        context.assertEqual(params[param_name], variable.params[param_name])
                    except KeyError:
                        context.fail('Expected variable parameter "%s" is absent' % param_name)

            if attrs:
                for attrs_name in attrs:
                    try:
                        context.assertEqual(attrs[attrs_name], getattr(variable, attrs_name))
                    except KeyError:
                        context.fail('Expected variable attribute "%s" is absent' % attrs_name)


class MockDali(ManagerMock, Dali):
    def __init__(self, bus):
        ManagerMock.__init__(self)
        Dali.__init__(self, 1, 2)
        self._send_interface = bus


class MockKnx(ManagerMock, Knx):
    def __init__(self, bus):
        ManagerMock.__init__(self)
        Knx.__init__(self, 1, 2)
        self._send_interface = bus


class MockLom(ManagerMock, Lom):
    def __init__(self, bus):
        ManagerMock.__init__(self)
        Lom.__init__(self, 1, 2)
        self._send_interface = bus
        self.devices={}


class MockBus:
    def __init__(self, responser=None):
        self.queue = deque()
        self.responser = responser

    def send(self, command, *args, **kwargs):
        self.queue.append(command)
        if self.responser:
            return self.responser(command, *args, **kwargs)

    def check_end_bus_messages(self, context: TestCase):
        if self.queue:
            unexpected_publishes = f'Unexpected bus message({len(self.queue)} pcs):\n'
            unexpected_publishes += "\n".join(str(i) for i in self.queue)
            context.fail(unexpected_publishes)

    def check_bus_message(self, context: TestCase, command):
        if len(self.queue) == 0:
            raise AssertionError('Expected command is absent: %s' % command)
        else:
            context.assertEqual(command, self.queue.popleft())

    def send_fast(self, command):
        print(command)

    @classmethod
    def of_dali(cls):
        def make_dali_response(command, *args, **kwargs):
            if command.is_query:
                return command._response(BackwardFrame(128))

        return cls(make_dali_response)

    @classmethod
    def of_knx(cls):
        def r(command, *args, state_address=None, **kwargs):
            yield 128

        return cls(r)

    @classmethod
    def of_lom(cls):
        def r():
            pass

        return cls(r)
