import os
import sys

from spread_core.mqtt import project
from spread_core.tests.topics.test_common import TestTopic


DEF_LOCAL = '/some/path/to/'


def build_path(root, *args):
    dest = []
    for part in args:
        dest += part.split('/')
    return os.path.join(root, *dest)


class TestLocalTopic(TestTopic):
    instance = project.Local

    def BROADCAST(self, *args):
        return self.instance.broadcast(*args)

    def OF_FILE(self, *args):
        return self.instance.of_file(*args)

    def PATH(self, *args):
        return self.instance.path(*args)


class TestPreset(TestLocalTopic):
    def setUp(self) -> None:
        self.instance = project.Preset(0, 'SomePreset', 1)

    def test_parsing(self):
        self.assertEqual(
            self.PARSE('Project/0/Local/Preset/SomePreset/1', project_id=0, entity_type='SomePreset', entity_id=1),
            self.instance
        )

    def test_broadcast(self):
        self.assertEqual(
            self.BROADCAST('Some1', 'Some2', 'Some3'),
            'Project/Some1/Local/Preset/Some2/Some3'
        )

    def test_file_interpretation(self):
        self.assertEqual(
            self.OF_FILE(0, 'Preset/SomePreset', '1.json'),
            self.instance
        )

    def test_file_generation(self):
        self.assertEqual(
            self.PATH(DEF_LOCAL),
            build_path(DEF_LOCAL, '0/Local/Preset/SomePreset/1.json')
        )


class TestCycle(TestLocalTopic):
    def setUp(self) -> None:
        self.instance = project.Cycle(0, 1)

    def test_parsing(self):
        self.assertEqual(
            self.PARSE('Project/0/Local/Cycle/1', project_id=0, entity_id=1),
            self.instance
        )

    def test_broadcast(self):
        self.assertEqual(
            self.BROADCAST('Some1', 'Some2'),
            'Project/Some1/Local/Cycle/Some2'
        )

    def test_file_interpretation(self):
        self.assertEqual(
            self.OF_FILE(0, 'Cycle', '1.json'),
            self.instance
        )

    def test_file_generation(self):
        self.assertEqual(
            self.PATH(DEF_LOCAL),
            build_path(DEF_LOCAL, '0/Local/Cycle/1')
        )


class TestCycleTrigger(TestLocalTopic):
    def setUp(self) -> None:
        self.instance = project.CycleTrigger(0, 1, 2)

    def test_parsing(self):
        self.assertEqual(
            self.PARSE('Project/0/Local/Cycle/1/Triggers/2', project_id=0, cycle_id=1, entity_id=2),
            self.instance
        )

    def test_broadcast(self):
        self.assertEqual(
            self.BROADCAST('Some1', 'Some2', 'Some3'),
            'Project/Some1/Local/Cycle/Some2/Triggers/Some3'
        )

    def test_file_interpretation(self):
        self.assertEqual(
            self.OF_FILE(0, 'Cycle/1/Triggers', '2.json'),
            self.instance
        )

    def test_file_generation(self):
        self.assertEqual(
            self.PATH(DEF_LOCAL),
            build_path(DEF_LOCAL, '0/Local/Cycle/1/Triggers/2.json')
        )


class TestEquipment(TestLocalTopic):
    def setUp(self) -> None:
        self.instance = project.Equipment(0, "Enginery", 1)

    def test_parsing(self):
        self.assertEqual(
            self.PARSE('Project/0/Local/Equipment/Enginery/1', project_id=0, entity_type='Enginery', entity_id=1),
            self.instance
        )

    def test_broadcast(self):
        self.assertEqual(
            self.BROADCAST('Some1', 'Some2', 'Some3'),
            'Project/Some1/Local/Equipment/Some2/Some3'
        )

    def test_file_interpretation(self):
        self.assertEqual(
            self.OF_FILE(0, 'Equipment/Enginery', '1.json'),
            self.instance
        )

    def test_file_generation(self):
        self.assertEqual(
            self.PATH(DEF_LOCAL),
            build_path(DEF_LOCAL, '0/Local/Equipment/Enginery/1.json')
        )


class TestLocation(TestLocalTopic):
    def setUp(self) -> None:
        self.instance = project.Location(0, 1)

    def test_parsing(self):
        self.assertEqual(
            self.PARSE('Project/0/Local/Location/1', project_id=0, entity_id=1),
            self.instance
        )

    def test_broadcast(self):
        self.assertEqual(
            self.BROADCAST('Some1', 'Some2'),
            'Project/Some1/Local/Location/Some2'
        )

    def test_file_interpretation(self):
        self.assertEqual(
            self.OF_FILE(0, 'Location', '1.json'),
            self.instance
        )

    def test_file_generation(self):
        self.assertEqual(
            self.PATH(DEF_LOCAL),
            build_path(DEF_LOCAL, '0/Local/Location/1.json')
        )


class TestSubginery(TestLocalTopic):
    def setUp(self) -> None:
        self.instance = project.Subginery(0, 1, "Subginery", 2)

    def test_parsing(self):
        self.assertEqual(
            self.PARSE('Project/0/Local/Location/1/Subginery/2', project_id=0, location_id=1, entity_type='Subginery', entity_id=2),
            self.instance
        )

    def test_broadcast(self):
        self.assertEqual(
            self.BROADCAST('Some1', 'Some2', 'Some3', 'Some4'),
            'Project/Some1/Local/Location/Some2/Some3/Some4'
        )

    def test_file_interpretation(self):
        self.assertEqual(
            self.OF_FILE(0, 'Location/1/Subginery', '2.json'),
            self.instance
        )

    def test_file_generation(self):
        self.assertEqual(
            self.PATH(DEF_LOCAL),
            build_path(DEF_LOCAL, '0/Local/Location/1/Subginery/2.json')
        )


class TestHardware(TestLocalTopic):
    def setUp(self) -> None:
        self.instance = project.Hardware(0, "Manager", 1)

    def test_parsing(self):
        self.assertEqual(
            self.PARSE('Project/0/Local/Hardware/Manager/1', project_id=0, entity_type='Manager', entity_id=1),
            self.instance
        )

    def test_broadcast(self):
        self.assertEqual(
            self.BROADCAST('Some1', 'Some2', 'Some3'),
            'Project/Some1/Local/Hardware/Some2/Some3'
        )

    def test_file_interpretation(self):
        self.assertEqual(
            self.OF_FILE(0, 'Hardware/Manager', '1.json'),
            self.instance
        )

    def test_file_generation(self):
        self.assertEqual(
            self.PATH(DEF_LOCAL),
            build_path(DEF_LOCAL, '0/Local/Hardware/Manager/1.json')
        )


class TestProvider(TestLocalTopic):
    def setUp(self) -> None:
        self.instance = project.Provider(0, "Manager", 1, "Provider", 2)

    def test_parsing(self):
        self.assertEqual(
            self.PARSE('Project/0/Local/Hardware/Manager/1/Provider/2', project_id=0, manager_type='Manager', manager_id=1, entity_type='Provider', entity_id=2),
            self.instance
        )

    def test_broadcast(self):
        self.assertEqual(
            self.BROADCAST('Some1', 'Some2', 'Some3', 'Some4', 'Some5'),
            'Project/Some1/Local/Hardware/Some2/Some3/Some4/Some5'
        )

    def test_file_interpretation(self):
        self.assertEqual(
            self.OF_FILE(0, 'Hardware/Manager/1/Provider', '2.json'),
            self.instance
        )

    def test_file_generation(self):

        self.assertEqual(
            self.PATH(DEF_LOCAL),
            build_path(DEF_LOCAL, '0/Local/Hardware/Manager/1/Provider/2.json')
        )


class TestHeader(TestLocalTopic):
    def setUp(self) -> None:
        self.instance = project.Header(0, "HeaderItem")

    def test_parsing(self):
        self.assertEqual(
            self.PARSE('Project/0/Local/Header/HeaderItem', project_id=0, entity_type='HeaderItem'),
            self.instance
        )

    def test_broadcast(self):
        self.assertEqual(
            self.BROADCAST('Some1', 'Some2'),
            'Project/Some1/Local/Header/Some2'
        )

    def test_file_interpretation(self):
        self.assertEqual(
            self.OF_FILE(0, 'Header', 'HeaderItem.json'),
            self.instance
        )

    def test_file_generation(self):

        self.assertEqual(
            self.PATH(DEF_LOCAL),
            build_path(DEF_LOCAL, '0/Local/Header/HeaderItem.json')
        )


class TestUser(TestLocalTopic):
    def setUp(self) -> None:
        self.instance = project.User(0, 1)

    def test_parsing(self):
        self.assertEqual(
            self.PARSE('Project/0/Local/User/1', project_id=0, entity_id=1),
            self.instance
        )

    def test_broadcast(self):
        self.assertEqual(
            self.BROADCAST('Some1', 'Some2'),
            'Project/Some1/Local/User/Some2'
        )

    def test_file_interpretation(self):
        self.assertEqual(
            self.OF_FILE(0, 'User', '1.json'),
            self.instance
        )

    def test_file_generation(self):
        self.assertEqual(
            self.PATH(DEF_LOCAL),
            build_path(DEF_LOCAL, '0/Local/User/1.json')
        )
