from spread_core.mqtt import TopicError, spread

from spread_core.tests.topics.test_common import TestTopic


class TestSpreadAddress(TestTopic):
    def parse_invalid_address(self, address):
        with self.assertRaises(TopicError):
            self.parse_address(address)

    @staticmethod
    def parse_address(address):
        return spread.address.of(*address.split('/'))

    def test_invalid_address(self):
        self.parse_invalid_address('Hardware/Provider/2/Funit')
        self.parse_invalid_address('1/qwe/Provider/2/Funit')

        self.parse_invalid_address('1/Hardware')
        self.parse_invalid_address('1/Hardware/Manager')
        self.parse_invalid_address('1/Hardware/Manager/2')

        self.parse_invalid_address('1/Hardware/Manager/2/Provider/3')
        self.parse_invalid_address('1/Hardware/Manager/2/Provider/3/Funit/qwe')

        self.parse_invalid_address('1/Equipment')
        self.parse_invalid_address('1/Equipment/Enginery')
        self.parse_invalid_address('1/Equipment/Enginery/2')
        self.parse_invalid_address('1/Equipment/Enginery/2/Funit/qwe')

        self.parse_invalid_address('1/Location')
        self.parse_invalid_address('1/Location/2')
        self.parse_invalid_address('1/Location/2/Subginery')
        self.parse_invalid_address('1/Location/2/Subginery/Funit/qwe')


class TestSpreadTopic(TestTopic):
    def test_invalid_topic(self):
        self.parse_invalid_topic('qwe')
        self.parse_invalid_topic('Spread')

        self.parse_invalid_topic('Spread/qwe/')
        self.parse_invalid_topic('Spread/State')
        self.parse_invalid_topic('Spread/Command')
        self.parse_invalid_topic('Spread/Reply')
        self.parse_invalid_topic('Spread/Reply/0')

    def test_reply_casting(self):
        self.assertEqual(
            spread.topic.Reply(0, 'some/session/'),
            'Spread/Reply/0/some/session/'
        )

    def test_raw_to_reply(self):
        topic = self.PARSE('Spread/Reply/0/some/session/')
        self.assertIsInstance(topic, spread.topic.Reply)
        self.assertEqual(topic.project_id, 0)
        self.assertEqual(topic.session_id, 'some/session/')

    def test_manager_casting(self):
        self.assertEqual(
            spread.topic.State(spread.address.ManagerAddress(0, 'Manager', 1, 'Funit')),
            'Spread/State/0/Hardware/Manager/1/Funit'
        )

    def test_manager(self):
        topic = self.PARSE('Spread/State/0/Hardware/Manager/1/Funit')
        self.assertIsInstance(topic, spread.topic.State)
        self.assertIsInstance(topic.entity_address,  spread.address.ManagerAddress)
        self.assertEqual(topic.entity_address.project_id, 0)
        self.assertEqual(topic.entity_address.type, 'Manager')
        self.assertEqual(topic.entity_address.id, 1)
        self.assertEqual(topic.entity_address.funit_type, 'Funit')

    def test_provider_casting(self):
        self.assertEqual(
            spread.topic.Set(spread.address.ProviderAddress(0, 'Manager', 1, 'Provider', 2, 'Funit')),
            'Spread/Set/0/Hardware/Manager/1/Provider/2/Funit'
        )

    def test_provider(self):
        topic = self.PARSE('Spread/State/0/Hardware/Manager/1/Provider/2/Funit')
        self.assertIsInstance(topic, spread.topic.State)
        self.assertIsInstance(topic.entity_address,  spread.address.ProviderAddress)
        self.assertEqual(topic.entity_address.project_id, 0)
        self.assertEqual(topic.entity_address.manager_type, 'Manager')
        self.assertEqual(topic.entity_address.manager_id, 1)
        self.assertEqual(topic.entity_address.type, 'Provider')
        self.assertEqual(topic.entity_address.id, 2)
        self.assertEqual(topic.entity_address.funit_type, 'Funit')

    def test_enginery_casting(self):
        self.assertEqual(
            spread.topic.Renew(spread.address.EngineryAddress(0, 'Enginery', 1, 'Funit')),
            'Spread/Renew/0/Equipment/Enginery/1/Funit'
        )

    def test_enginery(self):
        topic = self.PARSE('Spread/State/0/Equipment/Enginery/1/Funit')
        self.assertIsInstance(topic, spread.topic.State)
        self.assertIsInstance(topic.entity_address,  spread.address.EngineryAddress)
        self.assertEqual(topic.entity_address.project_id, 0)
        self.assertEqual(topic.entity_address.type, 'Enginery')
        self.assertEqual(topic.entity_address.id, 1)
        self.assertEqual(topic.entity_address.funit_type, 'Funit')

    def test_subginery_casting(self):
        self.assertEqual(
            spread.topic.State(spread.address.SubgineryAddress(0, 1, 'Subginery', 'Funit')),
            'Spread/State/0/Location/1/Subginery/Funit'
        )

    def test_subginery(self):
        topic = self.PARSE('Spread/State/0/Location/1/Subginery/Funit')
        self.assertIsInstance(topic, spread.topic.State)
        self.assertIsInstance(topic.entity_address,  spread.address.SubgineryAddress)
        self.assertEqual(topic.entity_address.project_id, 0)
        self.assertEqual(topic.entity_address.type, 'Subginery')
        self.assertEqual(topic.entity_address.id, 1)
        self.assertEqual(topic.entity_address.funit_type, 'Funit')