import datetime
import json
import uuid

import paho.mqtt.client as paho

from django import http
from django.forms.models import model_to_dict
from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from rest_framework.decorators import api_view
from rest_framework.response import Response

from spread_core import mqtt
from spread_core.bam.subgineries.incubator import Incubator
from spread_core.mqtt.spread.address import SubgineryAddress
from spread_core.tools.settings_const import CYCLES_FROM_0
from spread_dashboard.project import bam
from spread_dashboard.project.configs import Common
from spread_dashboard.project.serializers import LocationSerializer
from spread_dashboard.utils import app

LOCATION_FIELDS = {'id', 'name', 'cycle_id', 'cycle__name', 'cycle__period', 'start_time', 'recursive', 'active',
                   'scheduled', 'enabled'}
CYCLE_FIELDS = {'id', 'name', 'period', 'duration', 'startLevel', 'triggers'}

PROJECT_ID = app().project_id

@login_required
def index(request):
    template = 'incubator/new/index.html'
    context = {}
    return render(request, template, context)


@login_required
def config_api(request):
    config = app()
    locations = bam.models.Location.objects.all().order_by('id').values(*LOCATION_FIELDS)
    locations = LocationSerializer(locations, many=True)
    locations = list(locations.data)
    cycles = bam.models.Cycle.objects.all().order_by('name').values()
    cycles = list(cycles)

    res = dict(
            start_from_0=config.settings.get(CYCLES_FROM_0, False), #TODO разобраться или удолить
            locations=locations,
            cycles=cycles,
            start_index=0 if config.settings.get(CYCLES_FROM_0, False) else 1 #TODO разобраться или удолить
        )
    resp = http.HttpResponse(content=json.dumps(res))
    return resp


# def _convert_triggers_from_db(triggers):
#     """ Конвертирует триггеры из месива БД в словарь """
#     trig_dict = {}
#     for item in triggers.values():
#         key = item['day']
#         if key not in trig_dict:
#             trig_dict[key] = []
#         tpl = (
#             item['time'],
#             item['level'],
#         )
#         trig_dict[key].append(tpl)
#     # отсортируем массивы в новом словаре по времени
#     for key, lst in trig_dict.items():
#         trig_dict[key] = lst.sort()
#     return trig_dict


@api_view(['POST'])
def update_location(request):
    resp = {'result': None, 'error': None}
    data = request.data
    id_ = data.pop('id')
    resp['error'] = _publish_location_data(id_, data)
    if not resp['error']:
        # если все опубликовано успешно - сохраним в БД
        bam.models.Location.objects.update_or_create(id=id_, defaults=data)

    return Response(resp)


@api_view(['POST'])
def update_cycle(request):
    resp = {'result': None, 'error': None}
    data = request.data
    id_ = data.pop('id', None) or None
    # Если id есть, значит цикл уже существует - сперва публикуем в брокер, потом обновим в БД, если нет ошибок
    # иначе - сперва схороним в БД, чтоб получить id, потом опубликуем, а в случае ошибки - удолим
    if id_:
        resp['error'] = _publish_cycle_data(id_, data)
        if not resp['error']:
            model, _ = bam.models.Cycle.objects.update_or_create(id=id_, defaults=data)  # получим id, если его нет
    else:
        model, _ = bam.models.Cycle.objects.update_or_create(id=id_, defaults=data)
        resp['error'] = _publish_cycle_data(model.id, data)
        if resp['error']:
            bam.models.Cycle.objects.get(id=model.id).delete()

    return Response(resp)


@api_view(['POST'])
def delete_cycle(request):
    resp = {'result': None, 'error': None}
    id_ = request.data['id']
    topic = mqtt.project.Reset.broadcast(PROJECT_ID, 'Cycle', str(id_))
    resp['error'] = _publish_to_mqtt((topic, {}))
    if not resp['error']:
        bam.models.Cycle.objects.get(id=int(id_)).delete()

    return Response(resp)


def _get_broker_creds():
    """ Получение данных для подключения к брокеру согласно роли юзера """
    #TODO запрос на АУФ
    return 'admin', 'broker'


def _publish_to_mqtt(*pairs):
    """ Публикация сообщения в топик
    @pair: [tuple, list] - кортеж / список вида (topic, payload)
    """
    mqtt_client = paho.Client(str(uuid.uuid4()))
    creds = _get_broker_creds()
    mqtt_client.username_pw_set(*creds)
    broker_config = Common().serialize().get('broker')
    host = broker_config.get('host', 'localhost')
    port = broker_config.get('port', 1883)
    try:
        mqtt_client.connect(host, port)
        for topic, payload in pairs:
            payload = json.dumps(payload)
            mqtt_client.publish(topic=topic, payload=payload, qos=1)
        mqtt_client.disconnect()
    except Exception as exc:
        return str(exc)
    return None


def _publish_cycle_data(cid, data):
    """ Вспомогательная функция публикации данных цикла в соответствующие топики """
    pubs = []
    topic = mqtt.project.Set.broadcast(PROJECT_ID, 'Cycle', str(cid))
    triggers = data.pop('triggers', {})  # заберем триггеры на время публикации прочих настроек цикла
    pubs.append((topic, data))
    for key, payload in triggers.items():
        topic = mqtt.project.Set.broadcast(
            PROJECT_ID,
            'Cycle',
            str(cid),
            'Triggers',
            str(key)
        )
        pubs.append((topic, payload))
    data['triggers'] = triggers  # вернем на место для сохранения в БД
    return _publish_to_mqtt(*pubs)


def _publish_location_data(loc_id, data):
    """ Вспомогательная функция публикации связи локации и цикла в соответствующие топики """
    funit_dict = {
        'cycle_id': 'Cycle',
        'start_time': 'StartTime',
        'active': 'Active',
        'recursive': 'Recursive'
    }
    pubs = []
    for key, funit in funit_dict.items():
        topic = mqtt.spread.topic.Set(
            SubgineryAddress(
                PROJECT_ID,
                loc_id,
                'Incubator',
                funit
            )
        )
        pubs.append(
            (str(topic), {'value': data[key]})
        )
    return _publish_to_mqtt(*pubs)


#TODO админские права
def publish_location(loc_id):
    """ Принудительная публикация данных указанной локации. Служебный метод """
    resp = {'result': None, 'error': None}
    data = bam.models.Location.objects.get(id=loc_id)
    data = model_to_dict(data, fields=['cycle_id', 'start_time', 'active', 'recursive'])
    # сконвертим дату в строку
    start_time = data.get('start_time') or datetime.date.today()
    data['start_time'] = start_time.strftime('%Y-%m-%d')
    resp['error'] = _publish_location_data(loc_id, data)
    return Response(resp)


#TODO админские права
def publish_cycle(cycle_id):
    """ Принудительная публикация данных указанного цикла. Служебный метод """
    resp = {'result': None, 'error': None}
    data = bam.models.Cycle.objects.get(id=cycle_id)
    data = model_to_dict(data)
    data.pop('id')
    resp['error'] = _publish_cycle_data(cycle_id, data)
    return Response(resp)


# TODO старый функционал - удолить
@login_required
def cycles(request):
    return render(request, 'incubator/cycles.html', dict(
        broker=bam.broker_auth(),
        cycles=mqtt.project.Cycle.broadcast(app().project_id, ''),
        triggers=mqtt.project.CycleTrigger.broadcast(app().project_id),
        local=mqtt.project.LOCAL,
        set=mqtt.project.SET,
        reset=mqtt.project.RESET,
        start_index=0 if app().settings.get(CYCLES_FROM_0, False) else 1
    ))

# TODO старый функционал - удолить
@login_required
def cycle(request, cid):
    return render(request, 'incubator/cycle.html', dict(
        start_index=0 if app().settings.get(CYCLES_FROM_0, False) else 1,
        broker=bam.broker_auth(),
        data=str(mqtt.project.Cycle(app().project_id, cid)),
        triggers=str(mqtt.project.CycleTrigger.broadcast(app().project_id, cid, '')),
        local=mqtt.project.LOCAL,
        set=mqtt.project.SET,
        reset=mqtt.project.RESET,
    ))

# TODO старый функционал - удолить
@login_required
def locations(request):
    return render(
        request, 'incubator/locations.html',
        dict(
            broker=bam.broker_auth(),
            start_from_0=app().settings.get(CYCLES_FROM_0, False),
            locations=mqtt.project.ProjectFile.locations(app().project_id),
            cycles=mqtt.project.Cycle.broadcast(app().project_id, ''),
            states=mqtt.spread.topic.State(
                mqtt.spread.address.SubgineryAddress.broadcast(app().project_id, subginery_type=Incubator.__name__)
            ),
            start_index=0 if app().settings.get(CYCLES_FROM_0, False) else 1
        )
    )
