import json
import os
import uuid

import paho.mqtt.client as paho

from django.utils.translation import ugettext_lazy as _
from channels.generic.websocket import JsonWebsocketConsumer

from spread_dashboard.project.configs import Common
from spread_core.tools.path import CONFIG_PATH

class BrokerConsumer(JsonWebsocketConsumer):

    def __init__(self):
        super().__init__()
        self._var_dict = {}  # словарь связей топиков с человекопонятными переменными
        self._mqtt_client = None
        self._mqtt_config = Common().serialize().get('broker')

    def connect(self):
        """ Обработчик события подключения клиента по веб-сокету """
        self.accept()
        # проверяем авторизацию и получаем кредитсы для авторизации в брокере или прикрываем сокет
        creds = self._get_broker_role()
        self._init_mqtt_client(*creds)

    def disconnect(self, code: int):
        """ Обработчик события отключения клиента от веб-сокета """
        print('disconected:', code)
        if self._mqtt_client:
            self._deinit_mqtt_client()

    def receive_json(self,  payload):
        """ Метод, обрабатывающий запросы от клиента (браузера) """
        self._subscribe_to_topics(payload)
        self._unsubscribe_from_topics(payload)
        self._publish_to_topics(payload)

    def _subscribe_to_topics(self, payload: dict):
        """ Достает из payload список топиков и подписывается на все топики с QoS=0 """
        topics = payload.get('subscriptions')
        if not topics:
            return
        if not isinstance(topics, (list, tuple)):
            self._send_error(_('Поле "subscriptions" должно содержать список объектов'))
            return
        try:
            var_dict = {obj['topic']: obj.get('return_name') or obj['topic'] for obj in topics}
            topics = [(obj['topic'], obj.get('qos', 0)) for obj in topics]
            self._var_dict.update(var_dict)
            self._mqtt_client.subscribe(topics)
        except Exception as exc:
            self._send_error(exc)

    def _unsubscribe_from_topics(self, payload: dict):
        """ Достает из payload список топиков и отписывается от них """
        topics = payload.get('unsubscriptions')
        if not topics:
            return
        if not isinstance(topics, (list, tuple)):
            self._send_error(_('Поле "unsubscriptions" должно содержать список топиков'))
            return
        try:
            self._mqtt_client.unsubscribe(topics)
        except Exception as exc:
            self._send_error(exc)
        for key in topics:
            if key in self._var_dict:
                del self._var_dict[key]

    def _publish_to_topics(self, payload: dict):
        """ Достает из payload словарь с топиками и сообщениями и публикует их """
        publics = payload.get('publications')
        if not publics:
            return
        if not isinstance(publics, (list, tuple)):
            self._send_error(_('Поле "publications" должно содержать список топиков'))
            return
        try:
            for public in publics:
                qos = public.get('qos') or 0
                if qos not in (0, 1, 2):
                    self._send_error(_('Значение qos может быть 0, 1 или 2'))
                retain = bool(public.get('retain')) or False
                payload = json.dumps(public['payload'])
                self._mqtt_client.publish(topic=public['topic'], payload=payload, qos=qos, retain=retain)
        except Exception as exc:
            self._send_error(exc)

    def _send_error(self, msg: object):
        """ Публикует в вебсокет сообщение об ошибке """
        response = {'error': str(msg)}
        self.send_json(response)

    def _get_broker_role(self) -> tuple:
        """ Выдает логин и пароль от брокера, согласно списку разрешений с сервиса auth """
        #TODO запрос на АУФ
        if not self.scope:
            self._send_error(_('Пожалуйста, авторизуйтесь'))
            self.close(1337)
        # TODO обработка ответа
        return self._mqtt_config.get('login'), self._mqtt_config.get('password')

    def _on_receive_mqtt_message(self, client, userdata, message):
        """ Обработчик сообщений из брокера публикует сообщение в веб-сокет """
        payload = json.loads(bytes.decode(message.payload))
        var = self._var_dict.get(message.topic) or message.topic
        self.send_json({var: payload})

    def _on_connect_mqtt(self, client, userdata, flags, rc):
        print('on_connect', flags, rc)

    def _on_subscribe_mqtt(self, client, userdata, mid, granted_qos):
        print('on_subscribe', mid, granted_qos)

    def _on_publish_mqtt(self, client, userdata, mid):
        print('on_publish', mid)

    def _init_mqtt_client(self, login: str, password: str):
        """ Инициализирует подключение к брокеру. Необходим payload запроса, чтоб проверить аутентификацию """
        if not self._mqtt_client:
            self._mqtt_client = paho.Client(str(uuid.uuid4()))
            self._mqtt_client.username_pw_set(login, password)
            self._mqtt_client.on_message = self._on_receive_mqtt_message
            self._mqtt_client.on_connect = self._on_connect_mqtt
            self._mqtt_client.on_subscribe = self._on_subscribe_mqtt
            self._mqtt_client.on_publish = self._on_publish_mqtt
        if not self._mqtt_client.is_connected():
            host = self._mqtt_config.get('host', 'localhost')
            port = self._mqtt_config.get('port', 1883)
            self._mqtt_client.connect_async(host, port)
            self._mqtt_client.loop_start()
            # подписываемся на все топики из _var_dict
            subs = [{'topic': topic, 'return_name': var} for topic, var in self._var_dict.items()]
            self._subscribe_to_topics({'subscriptions': subs})

    def _deinit_mqtt_client(self):
        self._mqtt_client.loop_stop()
        unsubs = list(self._var_dict.keys())
        self._mqtt_client.unsubscribe(unsubs)
        self._mqtt_client.disconnect()
