import json
import logging
import os
import uuid
import zipfile

import spread_core.mqtt as mqtt
from spread_core.bam.animeo.manager import Animeo
from spread_core.bam.bolid.manager import Bolid
from spread_core.bam.dali.managers import Dali
from spread_core.bam.knx.manager import Knx
from spread_core.bam.lom.manager import Lom
from spread_core.bam.mercury.managers import Mercury
from spread_core.bam.subgineries import shading, lighting, alarming, handling, climating
from spread_core.mqtt import project
from spread_dashboard.project import models
from spread_dashboard.utils import app
from . import configs


def use_ctp(fn):
    def reader(*args, ctp=None, **kwargs):
        need_close = False
        if ctp is None:
            need_close = True
            ctp = zipfile.ZipFile(configs.Projector.ctp_file, 'r')

        try:
            return fn(*args, **kwargs, ctp=ctp)
        finally:
            if need_close:
                ctp.close()

    return reader


class Project:
    def __init__(self):
        ctp, project_id, root_location_id, self.name = self.read_pid()
        app().project_id = project_id
        configs.Config.PROJECT_ID = project_id
        if ctp:
            configs.Frontier.SERVER_ID = self.read_sid(ctp)

    @staticmethod
    def read_pid(ctp_path=None):
        """
        :param ctp_path: project.ctp path
        :return: ctp, project_id, root_location_id, name
        """
        if ctp_path is None:
            ctp_path = configs.Projector.ctp_file
        if os.path.exists(ctp_path):
            ctp = zipfile.ZipFile(ctp_path, 'r')
            entity = json.loads(ctp.read(project.ProjectFile.HEADER + '.json').decode())[project.ProjectFile.HEADER]
            return (
                ctp,
                entity['project']['id'],
                entity['project']['rootLocationID'],
                entity['project'].get('title', entity['project'].get('name', 'Project %s' % entity['project']['id']))
            )
        return None, None, None, None

    @staticmethod
    def read_sid(ctp):
        for entity in json.loads(ctp.read(project.ProjectFile.SERVERS + '.json').decode())[project.ProjectFile.SERVERS]:
            if entity['type'] == 'Spread':
                return entity['id']

    @classmethod
    def read_ctp(cls):
        with zipfile.ZipFile(configs.Projector.ctp_file, 'r') as ctp:
            cls.read_header(ctp=ctp)
            cls.read_ctp_file(models.Location, project.ProjectFile.LOCATIONS, ctp=ctp)
            cls.read_ctp_file(models.Subginery, project.ProjectFile.SUBGINERIES, ctp=ctp)
            cls.read_ctp_file(models.Manager, project.ProjectFile.MANAGERS, ctp=ctp)
            cls.read_ctp_file(models.Enginery, project.ProjectFile.ENGINERIES, ctp=ctp)
            cls.read_ctp_file(models.Provider, project.ProjectFile.PROVIDERS, ctp=ctp)
        models.RecipeIngredient.fill_recipes()
        models.Cycle.fill()

    @staticmethod
    @use_ctp
    def read_ctp_file(model, file, ctp):
        stack = []
        with ctp.open(file + '.json') as fs:
            for entity in json.load(fs)[file]:
                try:
                    assert issubclass(model, models.Entity)
                    # костыль, чтоб заполнить Provider
                    if model == models.Provider:
                        entity['manager_id'] = entity['managerID']
                        del entity['managerID']
                    model.objects.update_or_create(id=entity['id'], defaults=model.separator(entity))
                    stack.append(entity['id'])
                except BaseException as ex:
                    logging.error('Filling error: %s' % ex)
        model.objects.exclude(id__in=stack).delete()

    @staticmethod
    @use_ctp
    def read_header(ctp=None):
        with ctp.open(project.ProjectFile.HEADER + '.json') as fs:
            try:
                header = json.load(fs)[project.ProjectFile.HEADER]
                models.Header.objects.update_or_create(defaults=header)
            except BaseException as ex:
                logging.error('Header reading error:')
                logging.exception(ex)

    @classmethod
    def generate_configs(cls, force=False):
        ctp, project_id, _, _ = cls.read_pid(None)

        if force or project_id != configs.Config.PROJECT_ID:
            configs.Tree.reset()

        configs.Config.PROJECT_ID = project_id
        configs.Frontier.SERVER_ID = cls.read_sid(ctp)

        tree = configs.Tree()

        configs.Common().save(generate_sv=True)
        configs.Dashboard().save(generate_sv=True)
        if configs.Frontier.exist():
            configs.Frontier().save(generate_sv=True)
        configs.Projector().save(generate_sv=True)
        configs.Scheduler().save(generate_sv=True)
        configs.Scripter().save(generate_sv=True)
        configs.Cycler().save(generate_sv=True)
        configs.EventsSqlite().save(generate_sv=True)

        equipments = [shading.Shading.__name__, lighting.Lighting.__name__, alarming.Alarm.__name__, handling.Handling.__name__, climating.Climate.__name__]
        for subginery in json.loads(ctp.read(project.ProjectFile.SUBGINERIES + '.json').decode())[project.ProjectFile.SUBGINERIES]:
            if not equipments:
                break

            s_type = subginery['type']
            if s_type in equipments:
                if s_type == lighting.Lighting.__name__:
                    equipment = configs.LightingEquipment()
                elif s_type == shading.Shading.__name__:
                    equipment = configs.ShadingEquipment()
                elif s_type == alarming.Alarm.__name__:
                    equipment = configs.AlarmEquipment()
                elif s_type == handling.Handling.__name__:
                    equipment = configs.HandlingEquipment()
                elif s_type == climating.Climate.__name__:
                    equipment = configs.ClimateEquipment()
                else:
                    continue

                equipment.save(generate_sv=True)
                tree.decrease(equipment.config_dest, equipment.file)
                equipments.remove(s_type)

        app().publish_web_token()

        adapters = []
        for manager in json.loads(ctp.read(project.ProjectFile.MANAGERS + '.json').decode())[project.ProjectFile.MANAGERS]:
            adapter = None
            if manager['type'] == Dali.__name__:
                bus = manager['attributes']['bus']
                m = configs.DaliManager(bus)
                if bus not in adapters:
                    adapters.append(bus)
                    adapter = configs.CanAdapter(bus)
            elif manager['type'] == Mercury.__name__:
                bus = manager['attributes']['bus']
                m = configs.MercuryManager(bus)
                if bus not in adapters:
                    adapters.append(bus)
                    adapter = configs.TcpAdapter(bus)
            elif manager['type'] == Bolid.__name__:
                bus = manager['attributes']['bus']
                m = configs.BolidManager(bus)
                # if bus not in adapters:
                #     adapters.append(bus)
                #     adapter = configs.TcpAdapter(bus)
            elif manager['type'] == Lom.__name__:
                m = configs.LomManager()
            elif manager['type'] == Knx.__name__:
                bus = manager['attributes']['bus']
                m = configs.KnxManager(bus=bus)
                if bus not in adapters:
                    adapter = configs.KnxNetIpAdapter(bus)
                    adapters.append(bus)
            elif manager['type'] == Animeo.__name__:
                m = configs.AnimeoManager()
            else:
                continue

            if adapter:
                adapter.save(generate_sv=True)
                tree.decrease(adapter.config_dest, adapter.file)
            m.managerID = manager['id']
            m.save(generate_sv=True)
            tree.decrease(m.config_dest, m.file)

        tree.clear()


class Settings(dict):
    def __init__(self):
        super(Settings, self).__init__()
        self.read()

    def read(self) -> None:
        self.update(configs.Dashboard().serialize())
        for file in self.pop('include', []):
            if os.path.exists(file):
                with open(file, 'r') as stream:
                    self.update(json.load(stream))


def broker_auth() -> dict:
    guid = str(uuid.uuid4())
    reply = str(mqtt.spread.topic.Reply(app().project_id, guid))
    return dict(web_port=2883, login='admin', password='broker', reply=reply, guid=guid)


def config(provider: models.Provider):
    guid = uuid.uuid4()

    return dict(broker=broker_auth(),
                project=dict(state=str(provider.state('+')),
                             get=str(provider.renew('{}')),
                             set=str(provider.set('{}')),
                             reply=str(mqtt.spread.topic.Reply(app().project_id, str(guid))),
                             session_id=str(guid),
                             entity_id=provider.id if provider else None))
