import json
import logging
import os
import sys
import threading

from django.db import models
from django.core import validators
from django.utils.translation import gettext as _
from django.utils.translation import ugettext_lazy
from jsonfield import JSONField

import spread_core.mqtt as mqtt
from spread_core.bam import engineries, handling, subgineries
from spread_core.bam.engineries import handle
from spread_core.tools.settings_const import RESTART_TIMEOUT, TASKS, CALENDAR
from spread_dashboard import utils
from spread_dashboard.project import bam
from spread_dashboard.utils import app


class SchBase(models.Model):
    commit_timer = None

    @staticmethod
    def get_dict() -> dict:
        return {
            subgineries.lighting.Lighting.__name__: dict(
                topic=lambda e_id, funit: mqtt.spread.topic.Set(mqtt.spread.address.SubgineryAddress(app().project_id, e_id, subgineries.lighting.Lighting.__name__, funit)),
                name=_('scheduler_models_loc_self'),
                commands=[
                    dict(values=[True], name=_('scheduler_models_lighting_on'), funit=subgineries.lighting.On.funit_type),
                    dict(values=[1], name=_('scheduler_models_lighting_scene_1'), funit=subgineries.lighting.SceneOn.funit_type),
                    dict(values=[2], name=_('scheduler_models_lighting_scene_2'), funit=subgineries.lighting.SceneOn.funit_type),
                    dict(values=[True], name=_('scheduler_models_lighting_off'), funit=subgineries.lighting.Off.funit_type),
                    dict(validator=dict(min=0, max=100), name=_('scheduler_models_lighting_brightness'), funit=subgineries.lighting.BrightnessLevel.funit_type),
                ]
            ),
            engineries.LightingArea.__name__: dict(
                topic=lambda e_id, funit: mqtt.spread.topic.Set(mqtt.spread.address.EngineryAddress(app().project_id, engineries.LightingArea.__name__, e_id, funit)),
                name=_('scheduler_models_las'),
                commands=[
                    dict(name=_('scheduler_models_la_off'), funit=engineries.lighting_area.On.funit_type, values=[False]),
                    dict(name=_('scheduler_models_la_on'), funit=engineries.lighting_area.On.funit_type, values=[True]),
                    dict(name=_('scheduler_models_la_profile_funit'), funit=engineries.lighting_area.CurrentProfile.funit_type, values=[
                        ['Profile1', _('scheduler_models_la_profile_value').replace('{0}', '1')],
                        ['Profile2', _('scheduler_models_la_profile_value').replace('{0}', '2')],
                        ['Profile3', _('scheduler_models_la_profile_value').replace('{0}', '3')],
                        ['Profile4', _('scheduler_models_la_profile_value').replace('{0}', '4')],
                        ['Profile5', _('scheduler_models_la_profile_value').replace('{0}', '5')],
                    ]),
                ]
            ),
            engineries.DimmingLight.__name__: dict(
                topic=lambda e_id, funit: mqtt.spread.topic.Set(mqtt.spread.address.EngineryAddress(app().project_id, engineries.DimmingLight.__name__, e_id, funit)),
                name=_('scheduler_models_dimmers'),
                commands=[
                    dict(name=_('scheduler_models_dimmer_on'), funit=engineries.light.On.funit_type, values=[True]),
                    dict(name=_('scheduler_models_dimmer_off'), funit=engineries.light.On.funit_type, values=[False]),
                    dict(name=_('scheduler_models_dimmer_brightness'), funit=engineries.light.BrightnessLevel.funit_type, validator=dict(min=0, max=100)),
                    dict(name=_('scheduler_models_dimmer_group_brightness'), funit=engineries.light.BrightnessLevel.funit_type, validator=dict(min=0, max=100)),
                ]
            ),
            engineries.TunableWhiteLight.__name__: dict(
                topic=lambda e_id, funit: mqtt.spread.topic.Set(mqtt.spread.address.EngineryAddress(app().project_id, engineries.TunableWhiteLight.__name__, e_id, funit)),
                name=_('scheduler_models_tws'),
                commands=[
                    dict(name=_('scheduler_models_tw_on'), funit=engineries.light.On.funit_type, values=[True]),
                    dict(name=_('scheduler_models_tw_off'), funit=engineries.light.On.funit_type, values=[False]),
                    dict(name=_('scheduler_models_tw_brightness'), funit=engineries.light.BrightnessLevel.funit_type, validator=dict(min=0, max=100)),
                    dict(name=_('scheduler_models_tw_group_brightness'), funit=engineries.light.GroupLevel.funit_type, validator=dict(min=0, max=100)),
                    dict(name=_('scheduler_models_tw_temperature'), funit=engineries.light.Temperature.funit_type, validator=dict(min=2000, max=8000)),
                ]
            ),
            engineries.SwitchingLight.__name__: dict(
                topic=lambda e_id, funit: mqtt.spread.topic.Set(mqtt.spread.address.EngineryAddress(app().project_id, engineries.SwitchingLight.__name__, e_id, funit)),
                name=_('scheduler_models_switcher'),
                commands=[
                    dict(name=_('scheduler_models_switcher_on'), funit=engineries.light.On.funit_type, values=[True]),
                    dict(name=_('scheduler_models_switcher_off'), funit=engineries.light.On.funit_type, values=[False]),
                ]
            ),
            handle.Scenario.__name__: dict(
                topic=lambda e_id, funit: mqtt.spread.topic.Set(mqtt.spread.address.EngineryAddress(app().project_id, handle.Scenario.__name__, e_id, funit)),
                name=_('scheduler_models_scenarios'),
                commands=[
                    dict(name=_('scheduler_models_scenarios_on'), funit=handle.On.funit_type, values=[True]),
                    dict(name=_('scheduler_models_scenarios_off'), funit=handle.On.funit_type, values=[False]),
                ]
            )
        }

    class Meta:
        abstract = True

    def save(self, *args, timed=True, **kwargs):
        super(SchBase, self).save(*args, **kwargs)
        if timed:
            self.start_timer()

    def delete(self, *args, timed=True, **kwargs):
        super(SchBase, self).delete(*args, **kwargs)
        if timed:
            self.start_timer()

    @staticmethod
    def start_timer():
        if 'fill' in sys.argv:
            return
        if SchBase.commit_timer:
            SchBase.commit_timer.cancel()
            SchBase.commit_timer = None

        SchBase.commit_timer = threading.Timer(app().settings[RESTART_TIMEOUT], function=SchBase.on_timer)
        SchBase.commit_timer.start()

    @staticmethod
    def on_timer():
        SchBase.commit_timer = None
        SchBase.dump()
        Holiday.dump()
        LightFactor.dump()
        CommissioningStrategy.dump()
        os.system(f'sudo supervisorctl restart spread_scheduler')

    @staticmethod
    def dump():
        utils.write_file(
            app().settings[TASKS],
            [{
                **utils.serialize(location),
                'triggers': [
                    {
                        **utils.serialize(trigger),
                        'actions': [
                            utils.serialize(action) for action in SchAction.objects.filter(trigger=trigger)
                        ]
                    } for trigger in SchTrigger.objects.filter(location=location)
                ]
            } for location in SchLocation.objects.all()]
        )

    @staticmethod
    def fill():
        file = bam.configs.Dashboard().tasks
        if os.path.exists(file):
            try:
                with open(file, 'r') as fs:
                    data = json.load(fs)
            except BaseException as ex:
                logging.error(ex)
            else:
                for _loc in data:
                    triggers_data = _loc.pop('triggers')
                    location, _ = SchLocation.objects.update_or_create(id=_loc['id'], defaults=_loc)
                    try:
                        bam.models.Location.objects.get(id=location.id)
                    except bam.models.Location.DoesNotExist:
                        location.delete()
                    else:
                        for _tr in triggers_data:
                            actions_data = _tr.pop('actions')
                            trigger, _ = SchTrigger.objects.update_or_create(id=_tr['id'],
                                                                             defaults={**_tr, 'location': location})
                            for _act in actions_data:
                                SchAction.objects.update_or_create(id=_act['id'], defaults={**_act, 'trigger': trigger})

        SchLocation.objects.filter(
            id__in=[location.id for location in bam.models.Location.objects.filter(scheduled=False)]).delete()

        for loc in bam.models.Location.objects.filter(scheduled=True):
            try:
                SchLocation.objects.update_or_create(id=loc.id, defaults={'name': loc.name})
            except:
                continue


class SchLocation(SchBase):
    id = models.AutoField(primary_key=True)
    enabled = models.BooleanField(default=False)
    name = models.CharField(max_length=255)


class SchTrigger(SchBase):
    id = models.AutoField(primary_key=True)
    type = models.CharField(max_length=255)#, choices=[('workday', 'Рабочий день'), ('holiday', 'Выходной')])
    location = models.ForeignKey(SchLocation, on_delete=models.CASCADE, serialize=False)
    time = models.TimeField()


class SchAction(SchBase):
    id = models.AutoField(primary_key=True)
    trigger = models.ForeignKey(SchTrigger, on_delete=models.CASCADE, serialize=False)
    type = models.CharField(max_length=255, default='broker', editable=False)
    topic = models.CharField(max_length=255)
    value = JSONField()

    @property
    def funit_type(self):
        return self.topic.split('/')[-1]

    @property
    def command_name(self):
        topic = mqtt.topic_of(self.topic)
        for cmd in self.get_dict()[topic.entity_address.type]['commands']:
            if cmd['funit'] == topic.entity_address.funit_type:
                return cmd['name']

    @property
    def value_name(self):
        topic = mqtt.topic_of(self.topic)
        for cmd in self.get_dict()[topic.entity_address.type]['commands']:
            if cmd['funit'] == topic.entity_address.funit_type:
                if 'values' not in cmd:
                    return self.value
                elif len(cmd['values']) == 1:
                    return ''
                else:
                    for v in cmd['values']:
                        if v[0] == self.value:
                            return v[1]

        return 'X3'

    def to_view_data(self):
        d = SchBase.get_dict()

        action = handling.actions.action_of_data(utils.serialize(self))
        assert isinstance(action, handling.BrokerAction)
        target = action.entity_address.type

        commands = d[target]['commands']
        for i in range(0, len(commands)):
            funit = action.entity_address.funit_type
            if commands[i]['funit'] == funit:
                value = action.value
                if 'values' in commands[i]:
                    if len(commands[i]['values']) == 1:
                        if commands[i]['values'][0] != value:
                            continue
                    elif len(commands[i]['values']) > 1:
                        for v in range(0, len(commands[i]['values'])):
                            if commands[i]['values'][v][0] == value:
                                value = v
                                break

                return dict(
                    id=action.id,
                    target=list(d.keys()).index(target),
                    action=i,
                    value=value,
                    eId=action.entity_address.id
                )
        return {}

    @staticmethod
    def from_view_data(e_id, tr_id, target, act_id, index, value):
        d = SchBase.get_dict()
        e_type = list(d.keys())[target]
        item = d[e_type]['commands'][index]
        topic = d[e_type]['topic'](e_id, item['funit'])

        try:
            value = json.loads(value)
        except:
            pass

        if 'validator' in item and isinstance(value, int):
            if value not in range(item['validator']['min'], item['validator']['max'] + 1):
                raise ValueError(_('scheduler_models_unavailable_value'))

        if 'values' in item:
            if len(item['values']) == 1:
                value = item['values'][0]
            else:
                if value not in [v[0] for v in item['values']]:
                    raise ValueError(_('scheduler_models_unavailable_value'))

        if act_id:
            model = SchAction.objects.get(id=act_id)
            model.topic = topic
            model.value = value
            model.save()
        else:
            trigger = SchTrigger.objects.get(id=tr_id)
            model = SchAction.objects.create(**dict(trigger=trigger, topic=topic, value=value))

        return model

    @staticmethod
    def entities(loc_id: int):
        d = SchBase.get_dict()

        return sorted([
            {
                'name': _('scheduler_models_loc_self'),
                'entityId': loc_id,
                'entityType': subgineries.lighting.Lighting.__name__,
                'entityTypeIndex': 0,
            },
            *[
                {
                    'name': e.name[:-1],
                    'entityId': e.id,
                    'entityType': e.type,
                    'entityTypeIndex': list(d.keys()).index(e.type),
                } for e in bam.models.Enginery.get_scheduled(loc_id)
                if e.type in d
            ]
        ], key=lambda o: (o['entityTypeIndex'], o['name']))


class Holiday(SchBase):
    date = models.DateField(unique=True)

    @staticmethod
    def dump():
        utils.write_file(
            app().settings[CALENDAR],
            [str(model.date) for model in Holiday.objects.all()]
        )

    @staticmethod
    def fill():
        file = bam.configs.Dashboard().calendar
        if os.path.exists(file):
            try:
                with open(file, 'r') as fs:
                    data = json.load(fs)
            except BaseException as ex:
                logging.error(ex)
            else:
                Holiday.objects.all().delete()
                for datum in data:
                    Holiday.objects.create(date=datum)


class LightFactor(SchBase):
    class Meta:
        unique_together = (('area', 'enginery'),)

    value = models.FloatField(default=0.0, validators=[validators.MinValueValidator(0.0)])
    enginery = models.OneToOneField(bam.models.Enginery, on_delete=models.CASCADE, unique=True)
    area = models.ForeignKey(bam.models.Enginery, on_delete=models.CASCADE, related_name='area')

    @property
    def current_luminosity(self):
        return self.enginery.state('CurrentLuminosity')

    @staticmethod
    def dump():
        utils.write_file(
            bam.configs.Dashboard().light_factors,
            [{'value': model.value, 'enginery_id': model.enginery_id, 'area_id': model.area_id}
                for model in LightFactor.objects.all()]
        )

    @staticmethod
    def fill():
        file = bam.configs.Dashboard().light_factors
        if os.path.exists(file):
            try:
                with open(file, 'r') as fs:
                    data = json.load(fs)
            except BaseException as ex:
                logging.error(ex)
            else:
                LightFactor.objects.all().delete()
                for dictionary in data:
                    enginery = dictionary['enginery_id']
                    area = dictionary['area_id']
                    if bam.models.Enginery.objects.filter(id=enginery) and bam.models.Enginery.objects.filter(id=area):
                        LightFactor.objects.create(**dictionary)



class CommissioningStrategy(SchBase):
    area = models.ForeignKey(bam.models.Enginery, on_delete=models.DO_NOTHING, unique=True)
    natural_light = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_natural_light'))
    lum_flux_maintaining = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_lum_flux_maintaining'))
    daylight_glare_control = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_daylight_glare_control'))
    individual_configuration = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_individual_configuration'))
    load_balancing = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_load_balancing'))
    light_tuning = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_light_tuning'))
    scenes_using = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_scenes_using'))
    lighting_on_request = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_lighting_on_request'))
    presence_control = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_presence_control'))
    basic_setup = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_basic_setup'))
    schedule = models.BooleanField(default=False, verbose_name=ugettext_lazy('strategy_schedule'))

    def strategies(self, applied: bool = True) -> dict:
        """ Возвращает словарь, значениями которого являются переводы стратегий
        :param applied: вернуть только примененные (True) / непримененные (False) в световой зоне стратегии
        :return: словарь вида <поле>:<локализованное значение> """
        return {field.name: field.verbose_name for field in self._meta.fields
                if getattr(self, field.name) is applied}


    @staticmethod
    def fill():
        file = bam.configs.Dashboard().area_strategies
        if os.path.exists(file):
            try:
                with open(file, 'r') as fs:
                    data = json.load(fs)
            except BaseException as ex:
                logging.error(ex)
            else:
                CommissioningStrategy.objects.all().delete()
                for dictionary in data:
                    area_id = dictionary['area_id']
                    if bam.models.Enginery.objects.filter(id=area_id):
                        CommissioningStrategy.objects.create(**dictionary)

    @staticmethod
    def dump():
        utils.write_file(
            bam.configs.Dashboard().area_strategies,
            [
                {
                    'area_id': model.area_id,
                    'natural_light': model.natural_light,
                    'lum_flux_maintaining': model.lum_flux_maintaining,
                    'daylight_glare_control': model.daylight_glare_control,
                    'individual_configuration': model.individual_configuration,
                    'load_balancing': model.load_balancing,
                    'light_tuning': model.light_tuning,
                    'scenes_using': model.scenes_using,
                    'lighting_on_request': model.lighting_on_request,
                    'presence_control': model.presence_control,
                    'basic_setup': model.basic_setup,
                    'schedule': model.schedule,
                }
                for model in CommissioningStrategy.objects.all()
            ]
        )
